from __future__ import annotations

import abc
from d810.core import typing

import ida_hexrays

from d810.core import Registrant, getLogger, typing
from d810.errors import D810Exception
from d810.expr.ast import AstNode, minsn_to_ast
from d810.hexrays.hexrays_formatters import format_minsn_t, maturity_to_string
from d810.optimizers.microcode.handler import OptimizationRule

if typing.TYPE_CHECKING:
    from d810.core import OptimizationStatistics

d810_logger = getLogger("D810")
optimizer_logger = getLogger("D810.optimizer")


class InstructionOptimizationRule(OptimizationRule, Registrant, abc.ABC):
    """Base class for *instruction*-level optimizations.

    This class is now marked as *abstract* so that it is skipped when
    collecting concrete rules for presentation in the GUI.
    """

    CATEGORY = "Instruction"

    def __init__(self):
        super().__init__()
        self.maturities = []

    @abc.abstractmethod
    def check_and_replace(self, blk, ins):
        """Return a replacement instruction if the rule matches, otherwise None."""


class GenericPatternRule(InstructionOptimizationRule):
    PATTERNS: list[AstNode] | None = None

    def __init__(self):
        super().__init__()
        self.pattern_candidates = [self.PATTERN] if self.PATTERN is not None else []
        if self.PATTERNS is not None:
            self.pattern_candidates += self.PATTERNS

    @abc.abstractmethod
    def check_candidate(self, candidate: AstNode) -> bool:
        """Return True if the candidate matches the rule, otherwise False."""

    @property
    @abc.abstractmethod
    def PATTERN(self) -> AstNode:
        """Return the pattern to match."""

    @property
    @abc.abstractmethod
    def REPLACEMENT_PATTERN(self) -> AstNode:
        """Return the replacement pattern."""

    def get_valid_candidates(self, instruction: ida_hexrays.minsn_t, stop_early=True):
        valid_candidates = []
        tmp = minsn_to_ast(instruction)
        if tmp is None:
            return []
        for candidate_pattern in self.pattern_candidates:
            if not candidate_pattern:
                continue
            # Use a read-only check first for structural matching (no mops copied)
            if not candidate_pattern.check_pattern_and_copy_mops(tmp, read_only=True):
                continue
            # If the read-only check passes, then we can create a mutable copy
            # and populate mops
            mutable_candidate = candidate_pattern.clone()
            if not mutable_candidate.check_pattern_and_copy_mops(tmp):
                continue
            # Check candidate AFTER mops are populated, since check_candidate
            # may need to access leaf mops (e.g., to check segment permissions)
            if not self.check_candidate(mutable_candidate):
                continue
            valid_candidates.append(mutable_candidate)
            if stop_early:
                return valid_candidates
        return []

    def get_replacement(self, candidate: AstNode) -> ida_hexrays.minsn_t | None:
        # REPLACEMENT_PATTERN is implemented as a @property (or a method
        # masquerading as an attribute) that builds a brand-new AstNode tree
        # every time it is accessed.
        #
        # the issue is `self.REPLACEMENT_PATTERN` creates new nodes *EVERY* time because it
        # is a property. When invoked, it then re-instantiates the objects, so it loses
        # anything that modified it before.
        repl_pat = self.REPLACEMENT_PATTERN
        if not repl_pat:
            if optimizer_logger.debug_on:
                optimizer_logger.debug(
                    "No replacement pattern for rule %s",
                    self.NAME,
                )
            return None
        is_ok = repl_pat.update_leafs_mop(candidate)
        if optimizer_logger.debug_on:
            optimizer_logger.debug(
                "Replacement pattern updated leaf mops OK?: %s",
                is_ok,
            )
        if not is_ok:
            return None
        if not candidate.ea:
            if optimizer_logger.debug_on:
                optimizer_logger.debug(
                    "No EA for candidate %s",
                    candidate,
                )
            return None
        new_ins = repl_pat.create_minsn(candidate.ea, candidate.dst_mop)
        if optimizer_logger.debug_on:
            optimizer_logger.debug(
                "Replacement instruction created: %s",
                format_minsn_t(new_ins),
            )
        return new_ins

    @typing.override
    def check_and_replace(
        self, blk: ida_hexrays.mblock_t, instruction: ida_hexrays.minsn_t
    ) -> ida_hexrays.minsn_t | None:
        valid_candidates = self.get_valid_candidates(instruction, stop_early=True)
        if len(valid_candidates) == 0:
            return None
        new_instruction = self.get_replacement(valid_candidates[0])
        return new_instruction

    @property
    def description(self):
        if self.DESCRIPTION is not None:
            return self.DESCRIPTION
        if (self.PATTERN is None) or (self.REPLACEMENT_PATTERN is None):
            return ""
        self.PATTERN.reset_mops()
        self.REPLACEMENT_PATTERN.reset_mops()
        return "{0} => {1}".format(self.PATTERN, self.REPLACEMENT_PATTERN)


T_Rule = typing.TypeVar("T_Rule", bound=InstructionOptimizationRule)


class InstructionOptimizer(Registrant, typing.Generic[T_Rule]):
    RULE_CLASSES: list[typing.Type[T_Rule]] = []
    NAME = None

    def __init__(
        self, maturities: list[int], stats: OptimizationStatistics, log_dir=None
    ):
        self.rules: set[T_Rule] = set()
        self.maturities = maturities
        self.log_dir = log_dir
        self.cur_maturity = ida_hexrays.MMAT_PREOPTIMIZED
        # Centralized statistics collector injected by the manager
        self.stats = stats

    def add_rule(self, rule: T_Rule) -> bool:
        """Add a rule to this optimizer if it matches RULE_CLASSES.

        Rules must inherit from one of the classes in RULE_CLASSES to be accepted.
        This ensures type safety and proper interface compliance.
        """
        # Check if rule inherits from one of RULE_CLASSES
        is_valid_rule_class = False
        for rule_class in self.RULE_CLASSES:
            if isinstance(rule, rule_class):
                is_valid_rule_class = True
                break

        if not is_valid_rule_class:
            return False

        if optimizer_logger.debug_on:
            optimizer_logger.debug("Adding rule %s", rule)
        if len(rule.maturities) == 0:
            rule.maturities = self.maturities
        self.rules.add(rule)
        return True

    def get_optimized_instruction(
        self,
        blk: ida_hexrays.mblock_t,
        ins: ida_hexrays.minsn_t,
        *,
        allowed_rule_names: frozenset[str] | None = None,
    ) -> ida_hexrays.minsn_t | None:
        # Fast opcode gate for chain rules to avoid work on unrelated instructions
        # Only applies to optimizers whose rules expose a "TARGET_OPCODES" set.
        try:
            target_opcodes = getattr(self, "_allowed_root_opcodes", None)
            if target_opcodes:
                if ins.opcode not in target_opcodes:
                    return None
        except Exception:
            pass
        if blk is not None:
            self.cur_maturity = blk.mba.maturity
        # Optimizer-level maturity gate: skip entire optimizer if current
        # maturity is not in this optimizer's allowed maturities.
        # Per-rule maturity checks (below) provide defense-in-depth.
        if self.cur_maturity not in self.maturities:
            return None
        for rule in self.rules:
            if allowed_rule_names is not None and rule.name not in allowed_rule_names:
                continue
            if self.cur_maturity not in rule.maturities:
                continue
            try:
                new_ins = rule.check_and_replace(blk, ins)
                if new_ins is not None:
                    if self.stats is not None:
                        # Use new API with actual rule object
                        self.stats.record_rule_fired(
                            rule=rule,
                            optimizer=self.name,
                            maturity=self.cur_maturity,
                        )
                    optimizer_logger.info(
                        "Rule %s matched in maturity %s:",
                        rule.name,
                        maturity_to_string(self.cur_maturity),
                    )
                    optimizer_logger.info("  orig: %s", format_minsn_t(ins))
                    optimizer_logger.info("  new : %s", format_minsn_t(new_ins))

                    return new_ins
            except RuntimeError as e:
                optimizer_logger.error(
                    "Runtime error during rule %s in maturity %s for instruction %s: %s",
                    rule,
                    maturity_to_string(self.cur_maturity),
                    format_minsn_t(ins),
                    e,
                )
            except D810Exception as e:
                optimizer_logger.error(
                    "D810Exception during rule %s in maturity %s for instruction %s: %s",
                    rule,
                    maturity_to_string(self.cur_maturity),
                    format_minsn_t(ins),
                    e,
                )
        return None

    @property
    def name(self):
        if self.NAME is not None:
            return self.NAME
        return self.__class__.__name__


# Note: VerifiableRule instances are registered in RULE_REGISTRY (d810.mba.rules)
# and injected into PatternOptimizer at construction time by InstructionOptimizerManager.
# This avoids duck typing and keeps the registration explicit.



"""Export microcode to JSON format.

Export the current function's microcode at a specific maturity level,
with optional pre-deobfuscation snapshot capability.
Available from pseudocode view.
"""
from __future__ import annotations

import json
from d810.core import typing

from d810.core.logging import getLogger
from d810.ui.actions.base import D810ActionHandler

logger = getLogger("D810.ui")

ida_funcs = None
ida_hexrays = None
ida_kernwin = None

# ---------------------------------------------------------------------------
# Qt imports -- optional, will fail gracefully if not in GUI mode
# ---------------------------------------------------------------------------
try:
    from d810.qt_shim import (
        QtWidgets,
        QDialog,
        QVBoxLayout,
        QHBoxLayout,
        QLabel,
        QComboBox,
        QPushButton,
        QLineEdit,
    )

    QT_AVAILABLE = True
except ImportError:
    QT_AVAILABLE = False

# Maturity level choices (name, IDA constant value, description)
MATURITY_CHOICES = [
    ("GENERATED", 0, "Immediately after generation"),
    ("PREOPTIMIZED", 1, "Pre-optimized microcode"),
    ("LOCOPT", 2, "After local optimizations"),
    ("CALLS", 3, "After call analysis"),
    ("GLBOPT1", 4, "After first global optimization"),
    ("GLBOPT2", 5, "After second global optimization"),
    ("GLBOPT3", 6, "After third global optimization"),
    ("LVARS", 7, "Final with local variables"),
]


class ExportMicrocodeDialog(QDialog if QT_AVAILABLE else object):  # type: ignore[misc]
    """Dialog for configuring microcode export options."""

    def __init__(
        self,
        func_name: str,
        parent=None,
        ida_kernwin_module: typing.Any | None = None,
    ):
        """Initialize the export microcode dialog.

        Args:
            func_name: Name of the function being exported (for default filename)
            parent: Parent widget
        """
        super().__init__(parent)
        self.func_name = func_name
        self._ida_kernwin = ida_kernwin_module
        self.setWindowTitle("Export Microcode")
        self.setup_ui()

    def setup_ui(self):
        """Set up the dialog UI."""
        layout = QVBoxLayout()

        # Maturity level selection
        maturity_layout = QHBoxLayout()
        maturity_layout.addWidget(QLabel("Maturity level:"))
        self.maturity_combo = QComboBox()
        for name, _, description in MATURITY_CHOICES:
            self.maturity_combo.addItem(f"{name} - {description}", name)
        # Default to LVARS (final)
        self.maturity_combo.setCurrentIndex(7)
        maturity_layout.addWidget(self.maturity_combo)
        layout.addLayout(maturity_layout)

        # Pre/post deobfuscation radio buttons
        deobf_layout = QHBoxLayout()
        deobf_layout.addWidget(QLabel("Capture:"))
        self.post_deobf_radio = QtWidgets.QRadioButton("Post-deobfuscation (d810ng applied)")
        self.post_deobf_radio.setChecked(True)
        self.pre_deobf_radio = QtWidgets.QRadioButton("Pre-deobfuscation (d810ng suppressed)")
        deobf_layout.addWidget(self.post_deobf_radio)
        deobf_layout.addWidget(self.pre_deobf_radio)
        layout.addLayout(deobf_layout)

        # Output file selection
        file_layout = QHBoxLayout()
        file_layout.addWidget(QLabel("Output file:"))
        self.file_edit = QLineEdit()
        self.file_edit.setText(f"{self.func_name}_microcode.json")
        file_layout.addWidget(self.file_edit)

        browse_btn = QPushButton("Browse...")
        browse_btn.clicked.connect(self.browse_file)
        file_layout.addWidget(browse_btn)
        layout.addLayout(file_layout)

        # Buttons
        button_layout = QHBoxLayout()
        ok_btn = QPushButton("OK")
        ok_btn.clicked.connect(self.accept)
        cancel_btn = QPushButton("Cancel")
        cancel_btn.clicked.connect(self.reject)
        button_layout.addWidget(ok_btn)
        button_layout.addWidget(cancel_btn)
        layout.addLayout(button_layout)

        self.setLayout(layout)

    def browse_file(self):
        """Show file browser dialog."""
        if self._ida_kernwin is None:
            return

        file_path = self._ida_kernwin.ask_file(1, self.file_edit.text(), "Save microcode as...")
        if file_path:
            self.file_edit.setText(file_path)

    def get_settings(self) -> dict:
        """Get the configured export settings.

        Returns:
            Dictionary with maturity_name, maturity_value, pre_deobfuscation, output_path
        """
        idx = self.maturity_combo.currentIndex()
        maturity_name, maturity_value, _ = MATURITY_CHOICES[idx]

        return {
            "maturity_name": maturity_name,
            "maturity_value": maturity_value,
            "pre_deobfuscation": self.pre_deobf_radio.isChecked(),
            "output_path": self.file_edit.text(),
        }


class ExportMicrocode(D810ActionHandler):
    """Export the current function's microcode as JSON."""

    ACTION_ID = "d810ng:export_microcode"
    ACTION_TEXT = "Export microcode..."
    ACTION_TOOLTIP = "Export function microcode at specific maturity level"
    SUPPORTED_VIEWS = frozenset({"pseudocode"})
    MENU_ORDER = 72
    REQUIRES_STARTED = False

    def execute(self, ctx: typing.Any) -> int:
        """Execute the export microcode action.

        Args:
            ctx: IDA action context

        Returns:
            1 on success, 0 on failure
        """
        ida_funcs_mod = self.ida_module("ida_funcs", ida_funcs)
        ida_hexrays_mod = self.ida_module("ida_hexrays", ida_hexrays)
        ida_kernwin_mod = self.ida_module("ida_kernwin", ida_kernwin)
        if not QT_AVAILABLE or ida_funcs_mod is None or ida_hexrays_mod is None or ida_kernwin_mod is None:
            return 0

        # Get current function EA from pseudocode context
        vdui = ida_hexrays_mod.get_widget_vdui(ctx.widget)
        if vdui is None:
            logger.warning("ExportMicrocode: not in pseudocode view")
            ida_kernwin_mod.warning("Not in pseudocode view")
            return 0

        func_ea = vdui.cfunc.entry_ea
        func_name = ida_funcs_mod.get_func_name(func_ea)
        if not func_name:
            func_name = f"sub_{func_ea:X}"

        # Show dialog
        dialog = ExportMicrocodeDialog(func_name, ida_kernwin_module=ida_kernwin_mod)
        if dialog.exec_() != QDialog.Accepted:
            logger.info("Export microcode cancelled by user")
            return 0

        settings = dialog.get_settings()
        output_path = settings["output_path"]
        if not output_path:
            ida_kernwin_mod.warning("No output file specified")
            return 0

        maturity_value = settings["maturity_value"]
        pre_deobfuscation = settings["pre_deobfuscation"]

        try:
            # Import microcode dump utility
            from d810.hexrays.microcode_dump import dump_function_microcode

            # If pre-deobfuscation requested and d810ng is running, suppress hooks
            if pre_deobfuscation and hasattr(self._state, "manager") and self._state.manager.started:
                from d810.manager import d810_hooks_suppressed

                with d810_hooks_suppressed(self._state.manager):
                    microcode_data = dump_function_microcode(func_ea, maturity=maturity_value)
            else:
                # Normal export (post-deobfuscation or d810ng not running)
                microcode_data = dump_function_microcode(func_ea, maturity=maturity_value)

            # Write JSON to file
            with open(output_path, "w", encoding="utf-8") as f:
                json.dump(microcode_data, f, indent=2, ensure_ascii=False)

            logger.info("Exported microcode to %s", output_path)
            ida_kernwin_mod.info(f"Microcode exported to:\n{output_path}")
            return 1

        except Exception as exc:
            logger.error("Failed to export microcode: %s", exc, exc_info=True)
            ida_kernwin_mod.warning(f"Failed to export microcode:\n{exc}")
            return 0

    def is_available(self, ctx: typing.Any) -> bool:
        """Check if action is available in current context.

        Args:
            ctx: IDA action context

        Returns:
            True if in pseudocode view
        """
        ida_hexrays_mod = self.ida_module("ida_hexrays", ida_hexrays)
        if ida_hexrays_mod is None:
            return False

        # Check if we're in pseudocode view
        vdui = ida_hexrays_mod.get_widget_vdui(ctx.widget)
        return vdui is not None



"""Block Merge Optimization Rule.

Merges artificially split basic blocks by removing redundant goto
instructions.  When an obfuscator (e.g. Hikari, OLLVM) splits a single
basic block into many tiny blocks connected by unconditional jumps, this
rule detects the pattern and NOPs the intermediate ``m_goto`` instructions
so that IDA's built-in optimizer can recombine them.

Algorithm
---------
For every block *blk* passed to :meth:`BlockMerger.optimize`:

1. Check that *blk* ends with ``m_goto``.
2. Check that *blk* has exactly **one** successor.
3. Check that the goto destination operand (``mop_b``) matches the
   successor's serial.
4. Check that the successor has exactly **one** predecessor (i.e. *blk*
   itself).
5. If all conditions hold, NOP the ``m_goto`` instruction.  IDA will
   merge the two blocks in the next ``optimize_local`` pass.

Ported from **the copycat project** ``block_merge_handler_t`` (``block_merge.cpp``).
"""
from __future__ import annotations

import ida_hexrays

from d810.core import getLogger
from d810.optimizers.microcode.flow.handler import FlowOptimizationRule

logger = getLogger("D810.optimizer")


class BlockMerger(FlowOptimizationRule):
    """Merge artificially split basic blocks.

    This rule operates per-block.  When a block satisfies all mergeability
    criteria with its sole successor, the trailing ``m_goto`` is NOPed so
    that IDA's ``optimize_local`` pass combines the two blocks.

    CFG Safety
    ----------
    The rule only *removes* an instruction (NOP) and never adds edges or
    blocks.  Combined with ``USES_DEFERRED_CFG = False`` and a
    conservative ``SAFE_MATURITIES`` declaration, this is safe at the
    listed maturities.
    """

    DESCRIPTION = (
        "Merges artificially split basic blocks by removing "
        "redundant goto instructions"
    )

    # We directly NOP the goto -- no deferred CFG modification needed.
    # The actual block merging is done by IDA's optimize_local.
    USES_DEFERRED_CFG = False
    SAFE_MATURITIES = [
        ida_hexrays.MMAT_CALLS,
        ida_hexrays.MMAT_GLBOPT1,
    ]

    # ------------------------------------------------------------------
    # Public API
    # ------------------------------------------------------------------

    def optimize(self, blk: ida_hexrays.mblock_t) -> int:
        """Check whether *blk* can be merged with its successor.

        A block is mergeable when:

        1. It ends in ``m_goto``.
        2. It has exactly one successor.
        3. The successor has exactly one predecessor (this block).
        4. The goto destination (``mop_b``) equals the successor's serial.

        When all conditions are met the ``m_goto`` is NOPed and ``1`` is
        returned.  Otherwise ``0`` is returned.
        """
        if not self._can_merge(blk):
            return 0

        # NOP the goto -- IDA will merge the blocks in optimize_local.
        logger.info(
            "BlockMerger: NOPing goto in block %d -> %d",
            blk.serial,
            blk.tail.d.b,
        )
        blk.make_nop(blk.tail)
        return 1

    # ------------------------------------------------------------------
    # Internal helpers
    # ------------------------------------------------------------------

    @staticmethod
    def _can_merge(blk: ida_hexrays.mblock_t) -> bool:
        """Return ``True`` if *blk* can be merged with its successor.

        Criteria (mirroring the copycat project ``can_merge`` + ``has_single_goto_succ``):

        * *blk* has a tail instruction whose opcode is ``m_goto``.
        * *blk* has exactly one successor.
        * The goto destination operand type is ``mop_b`` and its value
          equals the successor's serial.
        * The successor has exactly one predecessor.
        """
        # --- tail must be m_goto ---
        if blk.tail is None:
            return False
        if blk.tail.opcode != ida_hexrays.m_goto:
            return False

        # --- single successor ---
        if blk.nsucc() != 1:
            return False

        succ_serial: int = blk.succ(0)

        # --- reject self-referencing goto (infinite loop guard) ---
        if succ_serial == blk.serial:
            return False

        # --- goto destination must reference the successor block ---
        if blk.tail.d.t != ida_hexrays.mop_b:
            return False
        if blk.tail.d.b != succ_serial:
            return False

        # --- successor must have exactly one predecessor ---
        mba = blk.mba
        succ_blk = mba.get_mblock(succ_serial)
        if succ_blk is None:
            return False
        if succ_blk.npred() != 1:
            return False

        return True



import collections
import dataclasses
import functools
import importlib
import sys
from abc import ABCMeta
from functools import cache, wraps
from types import GenericAlias, MappingProxyType
from collections.abc import MutableMapping
from weakref import WeakKeyDictionary

from d810.core.typing import (
    Annotated,
    Any,
    AnyStr,
    AsyncGenerator,
    Callable,
    ClassVar,
    Coroutine,
    ForwardRef,
    Generator,
    Generic,
    Hashable,
    Iterable,
    Literal,
    LiteralString,
    Optional,
    Self,
    Sequence,
    TypeAlias,
    TypeAliasType,
    TypeVar,
    cast,
    get_args,
    get_origin,
    get_type_hints,
    overload,
)

T = TypeVar("T")
_R = TypeVar("_R", bound="Registrant")
AnnotatedAny: TypeAlias = Annotated[Any, ...]  # safely parameterized


class NotGiven:
    """Placeholder for value which isn't given."""

    def __init__(self):
        raise NotImplementedError()

    def __bool__(self):
        return False

    def __repr__(self):
        return "NOT_GIVEN"

    @staticmethod
    def params(**kwargs):
        """Return a dict of the given parameters which are not NOT_GIVEN."""
        return {k: v for k, v in kwargs.items() if not isinstance(v, NotGiven)}


# Using __new__ to implement singleton pattern
NOT_GIVEN = object.__new__(NotGiven)
"""Placeholder for value which isn't given."""

# Pyright does not follow TypeVarType when passing type[T] to a function...
timestamp: TypeAlias = int
Thunk: TypeAlias = Callable[[], T]
OnePlus: TypeAlias = T | Sequence[T]
Defer: TypeAlias = T | Thunk[T]
TypeRef: TypeAlias = str | ForwardRef | GenericAlias | TypeAliasType | AnnotatedAny
DeferTypeRef: TypeAlias = Defer[type] | TypeRef
"""A typelike reference which can be wrapped to be resolved later."""


def survives_reload(cls=None, *, reload_key: str = ""):
    """
    Class decorator (optionally parameterized) that enables a class to survive reloads by storing a shared instance
    on the module object, keyed by `reload_key` (or `_SHARED_<ClassName>`).

    Usage:
        @survives_reload
        class MyClass: ...

        @survives_reload()
        class MyClass: ...

        BLAH = survives_reload(MyClass, reload_key="FOO")
    """
    def decorator(inner_cls):
        _reload_key = reload_key or f"_SHARED_{inner_cls.__name__}"
        _module = sys.modules[inner_cls.__module__]

        @functools.wraps(inner_cls)
        def get_shared_instance(*args, **kwargs):
            existing = getattr(_module, _reload_key, None)
            if existing is not None:
                return existing
            inst = inner_cls.__new__(inner_cls, *args, **kwargs)
            inner_cls.__init__(inst, *args, **kwargs)
            setattr(_module, _reload_key, inst)
            return inst

        return get_shared_instance

    # Handle all three cases:
    # 1. @survives_reload
    # 2. @survives_reload()
    # 3. survives_reload(SomeClass, reload_key=...)
    if cls is not None and callable(cls):
        return decorator(cls)
    return decorator


class CombineMeta:
    def __prepare__(
        self,
        name: str,
        bases: tuple[type, ...],
        **kwargs: Any
    ) -> MutableMapping[str, Any]:
        namespace: MutableMapping[str, Any] = {}
        for metaclass in self._get_most_derived_metaclasses(bases):
            ns = metaclass.__prepare__(name, bases, **kwargs)
            if type(ns) in (dict, type(namespace)):
                namespace.update(ns)
            else:
                if type(namespace) is not dict:
                    raise TypeError(
                        "metaclass conflict: " "multiple custom namespaces defined."
                    )
                ns.update(namespace)
                namespace = ns
        return namespace

    def __call__(
        self,
        name: str,
        bases: tuple[type, ...],
        namespace: MutableMapping[str, Any],
        **kwargs: Any
    ) -> type:
        metaclasses = self._get_most_derived_metaclasses(bases)
        if len(metaclasses) > 1:
            merged_name = "__".join(meta.__name__ for meta in metaclasses)
            ns = self.__prepare__(merged_name, tuple(metaclasses))
            metaclass = self(merged_name, tuple(metaclasses), ns, **kwargs)
        else:
            (metaclass,) = metaclasses or (type,)
        return metaclass(name, bases, dict(namespace), **kwargs)

    @staticmethod
    def _get_most_derived_metaclasses(
        bases: tuple[type, ...]
    ) -> list[type]:
        metaclasses: list[type] = []
        for metaclass in map(type, bases):
            if metaclass is not type:
                metaclasses = [
                    other for other in metaclasses if not issubclass(metaclass, other)
                ]
                if not any(issubclass(other, metaclass) for other in metaclasses):
                    metaclasses.append(metaclass)
        return metaclasses


combine_meta = CombineMeta()

def async_await(
    fn: Callable[..., Coroutine[Any, Any, T]],
) -> Callable[..., Generator[Any, None, T]]:
    """
    Decorator to convert an async function to a generator for use with a
    more intuitive async __await__.
    """

    @wraps(fn)
    def wrapper(*args, **kwargs):
        async_gen = fn(*args, **kwargs)
        return async_gen.__await__()

    return wrapper


def coroutine(
    fn: Callable[..., AsyncGenerator[None, T]],
) -> Callable[..., Coroutine[Any, Any, AsyncGenerator[None, T]]]:
    """Auto-starting coroutine decorator."""

    @wraps(fn)
    async def calls_asend(*args, **kwargs):
        gen = fn(*args, **kwargs)
        try:
            await gen.asend(None)  # type: ignore
        except StopAsyncIteration:
            print("StopAsyncIteration")  # Doesn't print
        return gen

    return calls_asend


def typename(t: TypeRef) -> str:
    """Return the name of a type, or the name of a value's type."""

    if get_origin(t) is None:
        if not isinstance(t, type):
            t = type(t)
        return t.__name__  # type: ignore
    return str(t)


def typecheck(value: Any, t: TypeRef) -> bool:
    """
    More featureful type checking. Supports isinstance, but also the zoo of
    `typing` types which are not supported by isinstance.
    """

    try:
        # type, Optional, Union, @runtime_checkable
        return isinstance(value, t)  # type: ignore
    except TypeError:
        pass

    if t is Any:
        return True
    if t in {None, type(None)}:
        return value is None
    if t in {AnyStr, LiteralString}:
        return isinstance(value, (str, bytes))

    # Generic types

    origin, args = get_origin(t), get_args(t)

    if origin is Literal:
        return value in args

    if origin is Annotated:
        return typecheck(value, args[0])

    if isinstance(t, TypeAliasType):
        return typecheck(value, t.__value__)  # type: ignore [attr-defined]

    return False


# Alternatively, could use ForwardRef._evaluate but that's private. This is at least public and legal.
def resolve_forward_ref(
    obj: TypeRef,
    globalns: dict[str, Any] | None = None,
    localns: dict[str, Any] | MappingProxyType[str, Any] | None = None,
):
    """Resolve a singular forward reference."""

    def dummy(x: TypeRef):
        pass

    _localns: dict[str, Any] = (
        cast(dict[str, Any], localns) if localns is not None else {}
    )
    return get_type_hints(dummy, globalns, _localns)["x"]


class deferred_property(Generic[T]):
    """A property which can be resolved later with minimal friction."""

    deferral: WeakKeyDictionary[type, Thunk[T]]

    def __init__(self):
        self.deferral = WeakKeyDictionary()

    def __set_name__(self, owner, name: str):
        self.__name__ = name

    @overload
    def __get__(self, instance: None, owner) -> Self: ...
    @overload
    def __get__(self, instance, owner) -> T: ...

    def __get__(self, instance, owner) -> Self | T:
        if instance is None:
            return self

        value = instance.__dict__.get(self.__name__, NOT_GIVEN)
        if value is not NOT_GIVEN:
            return value

        try:
            value = self.deferral.pop(instance)()
            setattr(instance, self.__name__, value)
            return value
        except KeyError:
            raise AttributeError(
                f"{typename(owner)}.{self.__name__} has no deferral"
            ) from None

    def __set__(self, instance, value: T):
        instance.__dict__[self.__name__] = value
        return value

    def defer(self, instance, deferral: Defer[T]):
        """Explicitly defer a value."""
        if callable(deferral):
            self.deferral[instance] = deferral
        else:
            setattr(instance, self.__name__, deferral)
        return self


def lazy_type(t: DeferTypeRef, cls: Optional[type] = None) -> Defer[type]:
    """Return a lazy type which can be resolved later."""

    if isinstance(t, (type, Callable)):
        return cast(type | Callable, t)
    if cls is None:
        raise ValueError("cls must be given if t is unbound")

    @cache
    def factory():
        global_ns = importlib.import_module(cls.__module__).__dict__
        return resolve_forward_ref(t, global_ns, cls.__dict__)

    return factory


class FilterableGenerator(Generic[T]):
    """
    Wraps an Iterable of classes and a list of predicates.
    You can .filter(...) repeatedly to build up predicates,
    and only when you iterate do we apply them.
    """

    def __init__(
        self,
        source: Iterable[T],
        predicates: list[Callable[[T], bool]] | None = None,
    ):
        self._source = source
        self._preds = list(predicates or [])

    def filter(self, predicate: Callable[[T], bool]) -> "FilterableGenerator[T]":
        return FilterableGenerator(self._source, self._preds + [predicate])

    def __iter__(self):
        for cls in self._source:
            if all(pred(cls) for pred in self._preds):
                yield cls

    def __repr__(self):
        # avoid consuming the generator!
        return f"<FilterableGenerator preds={len(self._preds)} source={self._source!r}>"


class Registry(ABCMeta):
    """Metaclass for registering subclasses."""

    def __init__(
        self,
        name: str,
        bases: tuple[type, ...],
        attrs: dict[str, Any],
    ):
        """Metaclass constructor that wires up class registries.

        Every concrete subclass receives its *own* ``registry`` and
        ``lazy_registry`` dictionaries so that sibling hierarchies do *not*
        accidentally share the same registry from a common ancestor.
        """

        super().__init__(name, bases, attrs)

        # Skip wiring for the Registrant sentinel itself.
        if name == "Registrant":
            return

        # If this is a direct subclass of Registrant, give it its own registries
        if Registrant in bases:
            self.registry: dict[str, type[Any]] = {}
            self.lazy_registry: dict[str, Thunk[type[Any]]] = {}


class Registrant(metaclass=Registry):
    """Self-registering resource."""

    registrant_name: ClassVar[str]
    """Name to register the resource under."""

    registry: ClassVar[dict[str, type[Any]]] = {}
    """Registry of registered resources."""

    lazy_registry: ClassVar[dict[str, Thunk[type[Any]]]] = {}
    """Registry of lazy registrations."""

    def __init_subclass__(cls):
        # Register *cls* into every immediate parent that is itself a
        # Registrant (except for the root Registrant, which we leave empty to
        # avoid an unwieldy global registry).
        visited = set()
        to_visit = list(cls.__bases__)
        while to_visit:
            base = to_visit.pop()
            if base in visited:
                continue
            visited.add(base)
            if Registrant in base.__bases__:
                base.register(cls)
            else:
                to_visit.extend(base.__bases__)

    @staticmethod
    def keyof(kls: type) -> str:
        """Return the key of the resource."""
        key_attr = getattr(kls, "registrant_name", kls.__name__)
        # # If someone defined `name` as a @property on the class we end up with a
        # # `property` object - not the actual value.  Fallback to class-name in
        # # that case.
        # if isinstance(key_attr, property):
        #     return cls.__name__
        return key_attr  # or cls.__name__

    @classmethod
    def normalize_key(cls, key: str) -> str:
        """Normalize a key."""
        return key.lower()

    @classmethod
    def register(cls, alt: type[Any]):
        """Directly register a subclass (unless it tries to register itself)."""
        # a class should not add itself to _its own_ registry
        if alt is cls and "registry" in cls.__dict__:
            return
        name = cls.normalize_key(cls.keyof(alt))
        # Pop any lazy registration
        cls.lazy_registry.pop(name, None)
        cls.registry[name] = alt

    @classmethod
    def lazy_register(cls, load: Thunk[type[Any]]):
        """Register a thunk (hook) under its function name for lazy initialization."""
        if load.__name__ not in cls.registry:
            cls.lazy_registry[load.__name__] = load

    @classmethod
    def get(cls, name: str) -> _R:  # type: ignore
        """Look up a registered subclass by name, loading lazily if needed."""
        key = cls.normalize_key(name)
        if factory := cls.lazy_registry.get(key):
            sub = factory()
            # move from lazy to real registry
            del cls.lazy_registry[key]
            cls.registry[key] = sub
            return cast(_R, sub)

        return cast(_R, cls.registry[key])

    @classmethod
    def find(cls, name: str) -> _R | None:  # type: ignore
        """Find a registered subclass by name (case-insensitive).

        Unlike get(), returns None if not found instead of raising KeyError.
        Useful for checking if a rule exists before removing it.
        """
        try:
            return cls.get(name)
        except KeyError:
            return None

    @classmethod
    def all(cls) -> list[type[Any]]:
        """Return every concrete subclass currently registered for *cls*."""
        return list(cls.registry.values())

    @classmethod
    def get_subclasses(cls, base: type | None = None) -> list[type]:
        """Return every concrete subclass of *base* that has been registered.

        Parameters
        ----------
        base: type
            The root class you are interested in (e.g. ``InstructionOptimizationRule``).

        Notes
        -----
        * ``base`` itself must ultimately inherit from :class:`Registrant`.
        * Works even when ``cls`` is the *Registrant* class itself.
        """

        if base is None:
            base = cls

        if not issubclass(base, Registrant):
            raise TypeError(
                f"get_subclasses() expects a Registrant-derived base class, received: {base} for class: {cls.__name__}"
            )

        # Every Registrant-derived class owns its *own* registry dict.  Simply
        # walk the MRO below *base* and aggregate them.
        collected: list[type] = []

        def _collect(sub):
            # Append concrete subclasses first
            if sub is not base and not getattr(sub, "__abstractmethods__", False):
                collected.extend(sub.registry.values())  # type: ignore[attr-defined]
            # Recurse into further subclasses
            for child in sub.__subclasses__():
                _collect(child)

        _collect(base)
        # Remove duplicates while preserving order
        unique: list[type] = []
        seen = set()
        for subcls in collected:
            if subcls not in seen:
                unique.append(subcls)
                seen.add(subcls)
        return unique

    @classmethod
    def filter(cls, predicate: Callable[[type], bool]) -> FilterableGenerator[type]:
        """
        Start a chain of filters over a generator of cls.registry.values().
        You can then .filter(...) again and again, and finally iterate it:
            for C in Registry.filter(p1).filter(p2): ...
        """
        # Exclude any classes earmarked for lazy loading so we don't trigger them
        # exclude: set[type] = set()
        # for load in cls.lazy_registry.values():
        #     try:
        #         exclude.add(load())
        #     except Exception:
        #         pass
        # gen = (c for c in cls.registry.values() if c not in exclude)
        return FilterableGenerator(cls.registry.values(), [predicate])


def get_all_subclasses(python_class: type) -> list[type]:
    """Return all subclasses of a class, recursively.

    Traverses the entire class hierarchy to find all concrete subclasses,
    returning them sorted by class name.
    """
    python_class.__subclasses__()

    subclasses = set()
    check_these = [python_class]

    while check_these:
        parent = check_these.pop()
        for child in parent.__subclasses__():
            if child not in subclasses:
                subclasses.add(child)
                check_these.append(child)

    return sorted(subclasses, key=lambda x: x.__name__)


class reify(Generic[T]):
    """
    Acts similar to a property, except the result will be
    set as an attribute on the instance instead of recomputed
    each access.
    """

    def __init__(self, fn: Callable[..., T]) -> None:
        self.fn = fn
        # Copy function attributes to preserve metadata
        self.__name__ = getattr(fn, "__name__", "<unknown>")
        self.__doc__ = getattr(fn, "__doc__", None)
        self.__module__ = getattr(fn, "__module__", "") or ""
        self.__qualname__ = getattr(fn, "__qualname__", "") or ""
        self.__annotations__ = getattr(fn, "__annotations__", {})

    @overload
    def __get__(self, instance: None, owner: type) -> "reify[T]": ...

    @overload
    def __get__(self, instance: Any, owner: type) -> T: ...

    def __get__(self, instance: Any, owner: type) -> "T | reify[T]":
        if instance is None:
            return self

        fn = self.fn
        val = fn(instance)
        setattr(instance, fn.__name__, val)
        return val


E = TypeVar("E", bound=Hashable)


@dataclasses.dataclass
class EventEmitter(Generic[E]):
    _listeners: collections.defaultdict[E, set[Callable]] = dataclasses.field(
        default_factory=lambda: collections.defaultdict(set), init=False
    )

    def on(self, event: E, handler: Callable | None = None):
        """Register an event handler for the given event."""
        if handler:
            self._listeners[event].add(handler)
            return handler

        @functools.wraps(self.on)
        def decorator(func):
            self.on(event, func)
            return func

        return decorator

    def once(self, event: E, handler: Callable):
        @functools.wraps(handler)
        def once_handler(*args, **kwargs):
            self.remove(event, once_handler)
            return handler(*args, **kwargs)

        self.on(event, once_handler)

    def remove(self, event: E, handler: Callable):
        self._listeners[event].discard(handler)

    def clear(self):
        self._listeners.clear()

    def emit(self, event: E, *args, **kwargs):
        for handler in self._listeners[event]:
            handler(*args, **kwargs)



"""Composable services for control-flow flattening removal.

This module decomposes the monolithic GenericDispatcherUnflatteningRule into
smaller, single-responsibility services that can be composed together.

Following the composition-over-inheritance principle, these services are:
- Easier to test in isolation
- Easier to understand (each does one thing)
- Easier to reuse in different contexts
- Easier to modify without breaking other functionality
"""

from __future__ import annotations

from dataclasses import dataclass, field
from d810.core.typing import List, Protocol, Tuple

import ida_hexrays

from d810.optimizers.core import OptimizationContext


@dataclass(frozen=True)
class Dispatcher:
    """Structured data representing a discovered control-flow flattening dispatcher.

    This immutable dataclass replaces the mutable GenericDispatcherInfo class,
    making dispatcher information explicit and easier to work with.

    A dispatcher is the central switch/dispatch block in a control-flow
    flattening obfuscation. It reads a state variable and jumps to different
    blocks based on that value.

    Attributes:
        entry_block: The microcode block that serves as the dispatcher entry.
        state_variable: The mop (microcode operand) used for dispatching decisions.
        comparison_values: List of constant values the state variable is compared against.
        internal_blocks: Blocks that are part of the dispatcher's internal logic.
        exit_blocks: Blocks that the dispatcher can transfer control to.
        mba: The microcode array containing this dispatcher.
    """
    entry_block: ida_hexrays.mblock_t
    state_variable: ida_hexrays.mop_t
    comparison_values: List[int] = field(default_factory=list)
    internal_blocks: List[ida_hexrays.mblock_t] = field(default_factory=list)
    exit_blocks: List[ida_hexrays.mblock_t] = field(default_factory=list)
    mba: ida_hexrays.mba_t | None = None

    def __str__(self) -> str:
        """Return a human-readable representation of this dispatcher."""
        return (
            f"Dispatcher(entry={self.entry_block.serial}, "
            f"state_var={self.state_variable}, "
            f"values={len(self.comparison_values)}, "
            f"exits={len(self.exit_blocks)})"
        )


class DispatcherFinder(Protocol):
    """Protocol for services that find dispatchers in microcode.

    Different obfuscation schemes (O-LLVM, Tigress, etc.) use different
    dispatcher patterns. By defining this as a protocol, we can have
    multiple implementations without coupling to a specific pattern.

    Example implementations:
    - OLLVMDispatcherFinder: Finds O-LLVM style dispatchers
    - TigressDispatcherFinder: Finds Tigress style dispatchers
    - GenericDispatcherFinder: Uses DispatcherCache for unknown obfuscators
    """

    def find(self, context: OptimizationContext) -> List[Dispatcher]:
        """Find all dispatchers in the given microcode.

        This method analyzes the control flow graph to identify blocks
        that act as dispatchers in a control-flow flattening scheme.

        Args:
            context: The optimization context containing the mba and configuration.

        Returns:
            A list of Dispatcher objects representing found dispatchers.
            Empty list if no dispatchers are found.

        Example:
            >>> finder = OLLVMDispatcherFinder()
            >>> dispatchers = finder.find(context)
            >>> for dispatcher in dispatchers:
            ...     print(f"Found dispatcher at block {dispatcher.entry_block.serial}")
        """
        ...


@dataclass(frozen=True)
class EmulationResult:
    """Result of emulating a dispatcher with tracked state values.

    Attributes:
        target_block: The target block reached after emulation, or None if failed.
        executed_instructions: Instructions executed during emulation.
        success: Whether emulation completed successfully.
        error_message: Error message if emulation failed.
    """
    target_block: ida_hexrays.mblock_t | None
    executed_instructions: List[ida_hexrays.minsn_t] = field(default_factory=list)
    success: bool = True
    error_message: str | None = None


class PathEmulator:
    """Emulates microcode execution paths to resolve state variables.

    This service wraps the complex MopTracker and MicroCodeInterpreter logic
    into a simple, testable interface. It's responsible for determining what
    value a state variable will have when reaching a dispatcher.

    The emulation is key to unflattening: we need to know which dispatcher
    exit block will be taken for each predecessor block.

    Example:
        >>> emulator = PathEmulator()
        >>> target = emulator.resolve_target(context, pred_block, dispatcher)
        >>> print(f"Block {pred_block.serial} will jump to {target.serial}")
    """

    def resolve_target(
        self,
        context: OptimizationContext,
        from_block: ida_hexrays.mblock_t,
        dispatcher: Dispatcher
    ) -> ida_hexrays.mblock_t | None:
        """Emulate execution from from_block through the dispatcher.

        This method:
        1. Tracks the dispatcher's state variable backwards from from_block
        2. Determines its constant value (if resolvable)
        3. Emulates the dispatcher logic with that value
        4. Returns the actual target block (bypassing the dispatcher)

        Args:
            context: The optimization context.
            from_block: The predecessor block jumping to the dispatcher.
            dispatcher: The dispatcher to emulate through.

        Returns:
            The actual target block that from_block should jump to directly.
            None if the target cannot be resolved (e.g., state variable depends on input).

        Raises:
            ValueError: If emulation fails unexpectedly.

        Example:
            >>> # Before: block_5 -> dispatcher -> block_42
            >>> # After:  block_5 -> block_42 (dispatcher bypassed)
            >>> target = emulator.resolve_target(context, block_5, dispatcher)
            >>> assert target == block_42
        """
        result = self.emulate_with_history(context, from_block, dispatcher)
        return result.target_block

    def emulate_with_history(
        self,
        context: OptimizationContext,
        from_block: ida_hexrays.mblock_t,
        dispatcher: Dispatcher,
        mop_history: "MopHistory | None" = None
    ) -> EmulationResult:
        """Emulate dispatcher execution with tracked state values.

        This is the core emulation method that wraps the existing
        MicroCodeInterpreter and MicroCodeEnvironment logic.

        Args:
            context: The optimization context.
            from_block: The predecessor block to emulate from.
            dispatcher: The dispatcher to emulate through.
            mop_history: Optional pre-computed MopHistory. If None, tracking
                        is performed automatically from from_block.

        Returns:
            EmulationResult containing the target block and execution details.
        """
        from d810.expr.emulator import MicroCodeEnvironment, MicroCodeInterpreter
        from d810.hexrays.helper import format_minsn_t, format_mop_t
        from d810.hexrays.tracker import MopHistory, MopTracker

        context.logger.debug(
            "Emulating dispatcher %s from block %s",
            dispatcher.entry_block.serial,
            from_block.serial
        )

        # Step 1: Get or create MopHistory by tracking from from_block
        if mop_history is None:
            if dispatcher.state_variable is None:
                return EmulationResult(
                    target_block=None,
                    success=False,
                    error_message="Dispatcher has no state variable"
                )
            # Track the state variable backwards from from_block
            tracker = MopTracker([dispatcher.state_variable], context.mba)
            mop_history = tracker.search_backward(from_block)
            if mop_history is None:
                return EmulationResult(
                    target_block=None,
                    success=False,
                    error_message=f"Could not track state variable from block {from_block.serial}"
                )

        # Step 2: Setup the microcode environment with state variable values
        microcode_interpreter = MicroCodeInterpreter(symbolic_mode=False)
        microcode_environment = MicroCodeEnvironment()

        # Get value of the state variable from the tracked history
        state_value = mop_history.get_mop_constant_value(dispatcher.state_variable)
        if state_value is None:
            return EmulationResult(
                target_block=None,
                success=False,
                error_message=f"State variable value not resolvable from block {from_block.serial}"
            )

        # Initialize the environment with the state variable value
        microcode_environment.define(dispatcher.state_variable, state_value)

        context.logger.debug(
            "Executing dispatcher %s with: %s = %x",
            dispatcher.entry_block.serial,
            format_mop_t(dispatcher.state_variable),
            state_value
        )

        # Step 3: Emulate the dispatcher blocks
        internal_block_serials = {blk.serial for blk in dispatcher.internal_blocks}
        all_dispatcher_serials = {dispatcher.entry_block.serial} | internal_block_serials

        instructions_executed = []
        cur_blk = dispatcher.entry_block
        cur_ins = cur_blk.head

        # Continue while in dispatcher blocks (not at an exit block)
        max_iterations = 1000  # Safety limit
        iteration = 0
        while (
            cur_blk is not None
            and cur_blk.serial in all_dispatcher_serials
            and iteration < max_iterations
        ):
            iteration += 1
            context.logger.debug(
                "  Executing: %s.%s",
                cur_blk.serial,
                format_minsn_t(cur_ins) if cur_ins else "None"
            )

            if cur_ins is None:
                break

            # Evaluate the current instruction
            is_ok = microcode_interpreter.eval_instruction(
                cur_blk, cur_ins, microcode_environment
            )
            if not is_ok:
                # Emulation could not continue - return current block
                return EmulationResult(
                    target_block=cur_blk,
                    executed_instructions=instructions_executed,
                    success=True
                )

            instructions_executed.append(cur_ins)
            cur_blk = microcode_environment.next_blk
            cur_ins = microcode_environment.next_ins

        if iteration >= max_iterations:
            return EmulationResult(
                target_block=None,
                executed_instructions=instructions_executed,
                success=False,
                error_message="Emulation exceeded maximum iterations"
            )

        # Return the first block that is not part of the dispatcher
        return EmulationResult(
            target_block=cur_blk,
            executed_instructions=instructions_executed,
            success=True
        )


class CFGPatcher:
    """Applies changes to the control-flow graph.

    This service encapsulates all CFG modification logic, making it easy to:
    - Test in isolation (mocking is straightforward)
    - Audit for correctness (all graph changes in one place)
    - Extend with new types of patches

    All methods are static because they don't need instance state - they
    operate purely on the provided mba and blocks.

    Implementation notes:
        This class wraps the low-level functions from d810.hexrays.cfg_utils,
        providing a clean interface that fits the composition-based architecture.
        The underlying implementations handle all the complex bookkeeping
        (predecessor/successor lists, block types, verification).
    """

    @staticmethod
    def redirect_edge(
        context: OptimizationContext,
        from_block: ida_hexrays.mblock_t,
        to_block: ida_hexrays.mblock_t,
        verify: bool = True
    ) -> int:
        """Redirect a block's outgoing edge to a new target.

        This is the fundamental operation for control-flow unflattening:
        changing a block that jumps to the dispatcher to instead jump
        directly to its actual target.

        Args:
            context: The optimization context.
            from_block: The block whose outgoing edge will be changed.
            to_block: The new target block.

        Returns:
            The number of changes made (typically 1 if successful, 0 if no-op).

        Raises:
            RuntimeError: If the CFG modification fails verification.

        Example:
            >>> # Unflatten: block_5 -> dispatcher -> block_42
            >>> # Becomes:   block_5 -> block_42
            >>> changes = CFGPatcher.redirect_edge(context, block_5, block_42)
        """
        from d810.hexrays.cfg_utils import (
            change_0way_block_successor,
            change_1way_block_successor,
            make_2way_block_goto,
        )

        context.logger.debug(
            "Redirecting block %s to %s",
            from_block.serial,
            to_block.serial
        )

        nsucc = from_block.nsucc()

        if nsucc == 0:
            # Block has no successors (e.g., ends with indirect jump or return)
            # We can still redirect by adding a goto
            success = change_0way_block_successor(from_block, to_block.serial, verify=verify)
        elif nsucc == 1:
            # Single successor - most common case
            success = change_1way_block_successor(from_block, to_block.serial, verify=verify)
        elif nsucc == 2:
            # Conditional block - convert to unconditional goto
            success = make_2way_block_goto(from_block, to_block.serial, verify=verify)
        else:
            context.logger.warning(
                "Cannot redirect block %s with %d successors",
                from_block.serial,
                nsucc
            )
            return 0

        return 1 if success else 0

    @staticmethod
    def insert_intermediate_block(
        context: OptimizationContext,
        before_block: ida_hexrays.mblock_t,
        after_block: ida_hexrays.mblock_t,
        instructions: List[ida_hexrays.minsn_t],
        verify: bool = True
    ) -> ida_hexrays.mblock_t | None:
        """Insert a new block between two existing blocks.

        Sometimes the dispatcher performs computations that need to be
        preserved when unflattening. This method creates a new block
        containing those instructions and inserts it in the CFG.

        Args:
            context: The optimization context.
            before_block: The block that will jump to the new block.
            after_block: The block that the new block will jump to.
            instructions: The instructions to place in the new block.

        Returns:
            The newly created block, or None if creation failed.

        Raises:
            RuntimeError: If the CFG modification fails verification.

        Example:
            >>> # The dispatcher does: x = x + 1; jump to block_42
            >>> # We need to preserve that computation
            >>> new_blk = CFGPatcher.insert_intermediate_block(
            ...     context, block_5, block_42, [add_instruction])
        """
        from d810.hexrays.cfg_utils import (
            change_1way_block_successor,
            create_block,
        )

        context.logger.debug(
            "Inserting intermediate block between %s and %s with %d instructions",
            before_block.serial,
            after_block.serial,
            len(instructions)
        )

        # Create a new block with the instructions
        # The new block will initially point to the next block after before_block
        new_block = create_block(before_block, instructions, is_0_way=False, verify=verify)

        if new_block is None:
            context.logger.error(
                "Failed to create intermediate block between %s and %s",
                before_block.serial,
                after_block.serial
            )
            return None

        # Redirect the new block to point to after_block
        success = change_1way_block_successor(new_block, after_block.serial, verify=verify)
        if not success:
            context.logger.error(
                "Failed to redirect new block %s to %s",
                new_block.serial,
                after_block.serial
            )
            return None

        context.logger.debug(
            "Created intermediate block %s between %s and %s",
            new_block.serial,
            before_block.serial,
            after_block.serial
        )

        return new_block

    @staticmethod
    def ensure_unconditional_predecessor(
        context: OptimizationContext,
        father_block: ida_hexrays.mblock_t,
        child_block: ida_hexrays.mblock_t,
        verify: bool = True
    ) -> int:
        """Ensure a predecessor block has an unconditional jump to child.

        Some optimizations require that predecessor blocks end with
        unconditional jumps (goto) rather than conditional jumps.
        This method transforms the CFG to meet that requirement by
        inserting an intermediate block if necessary.

        Args:
            context: The optimization context.
            father_block: The predecessor block to check/modify.
            child_block: The child block that father should jump to unconditionally.

        Returns:
            The number of changes made (0 if already unconditional, 1 if modified).

        Raises:
            RuntimeError: If the CFG modification fails verification.

        Example:
            >>> # Before: block_5 ends with conditional jump to dispatcher
            >>> # After:  block_5 -> new_block (unconditional) -> dispatcher
            >>> changes = CFGPatcher.ensure_unconditional_predecessor(
            ...     context, block_5, dispatcher_block)
        """
        from d810.hexrays.cfg_utils import ensure_child_has_an_unconditional_father

        if father_block is None:
            return 0

        context.logger.debug(
            "Ensuring block %s has unconditional jump to block %s",
            father_block.serial,
            child_block.serial
        )

        return ensure_child_has_an_unconditional_father(father_block, child_block, verify=verify)

    @staticmethod
    def duplicate_block(
        context: OptimizationContext,
        block: ida_hexrays.mblock_t,
        verify: bool = True
    ) -> Tuple[ida_hexrays.mblock_t, ida_hexrays.mblock_t | None]:
        """Duplicate a block in the CFG.

        Creates a copy of a block at the end of the block array. If the
        original block ends with a conditional jump, also creates a
        default successor block.

        Args:
            context: The optimization context.
            block: The block to duplicate.

        Returns:
            A tuple of (duplicated_block, default_block).
            default_block is None if the original wasn't conditional.

        Raises:
            RuntimeError: If the CFG modification fails verification.

        Example:
            >>> dup, default = CFGPatcher.duplicate_block(context, dispatcher_block)
            >>> # dup is the new copy of dispatcher_block
        """
        from d810.hexrays.cfg_utils import duplicate_block

        context.logger.debug("Duplicating block %s", block.serial)

        dup_block, dup_default = duplicate_block(block, verify=verify)

        context.logger.debug(
            "Duplicated block %s -> %s (default: %s)",
            block.serial,
            dup_block.serial,
            dup_default.serial if dup_default else None
        )

        return dup_block, dup_default

    @staticmethod
    def clean_cfg(
        context: OptimizationContext,
        mba: ida_hexrays.mba_t,
        merge_blocks: bool = False
    ) -> int:
        """Clean up the CFG by removing empty blocks and simple gotos.

        After unflattening, the CFG may contain unnecessary blocks
        (empty blocks, simple goto chains). This method cleans them up.

        Args:
            context: The optimization context.
            mba: The microcode block array to clean.
            merge_blocks: Whether to merge consecutive blocks (can be unstable).

        Returns:
            The number of changes made.

        Note:
            The merge_blocks option calls mba.merge_blocks() which can cause
            crashes in some scenarios. Use with caution.

        Example:
            >>> changes = CFGPatcher.clean_cfg(context, mba)
        """
        from d810.hexrays.cfg_utils import mba_deep_cleaning

        context.logger.debug(
            "Cleaning CFG (merge_blocks=%s, maturity=%s)",
            merge_blocks,
            mba.maturity
        )

        return mba_deep_cleaning(mba, call_mba_combine_block=merge_blocks)


class OLLVMDispatcherFinder:
    """Finds O-LLVM style dispatchers in microcode.

    This service implements the DispatcherFinder protocol for O-LLVM style
    control-flow flattening. O-LLVM dispatchers have distinctive characteristics:

    - A central "dispatcher" block with many predecessors
    - State variable compared against multiple constant values
    - High entropy in comparison values (pseudo-random)
    - Internal blocks that only compute next state + jump back

    This implementation wraps the existing OllvmDispatcherInfo class using
    the Strangler Fig pattern, providing a clean interface while reusing
    proven logic.

    Example:
        >>> finder = OLLVMDispatcherFinder()
        >>> dispatchers = finder.find(context)
        >>> for d in dispatchers:
        ...     print(f"Found: {d}")
    """

    # Entropy thresholds for O-LLVM detection
    # Values are pseudo-random, so entropy should be ~0.5
    DEFAULT_MIN_ENTROPY = 0.3
    DEFAULT_MAX_ENTROPY = 0.7

    def __init__(
        self,
        min_entropy: float | None = None,
        max_entropy: float | None = None,
        use_dispatcher_cache: bool = True
    ):
        """Initialize the finder with optional configuration.

        Args:
            min_entropy: Minimum entropy threshold (default 0.3).
            max_entropy: Maximum entropy threshold (default 0.7).
            use_dispatcher_cache: Whether to use DispatcherCache for early filtering.
        """
        self._min_entropy = min_entropy or self.DEFAULT_MIN_ENTROPY
        self._max_entropy = max_entropy or self.DEFAULT_MAX_ENTROPY
        self._use_dispatcher_cache = use_dispatcher_cache

    def find(self, context: OptimizationContext) -> List[Dispatcher]:
        """Find all O-LLVM dispatchers in the given microcode.

        This method:
        1. Uses DispatcherCache to filter candidate blocks (if enabled)
        2. Guesses the outermost dispatcher using predecessor count
        3. Explores each candidate to validate it's a real dispatcher
        4. Converts validated dispatchers to immutable Dispatcher objects

        Args:
            context: The optimization context containing the mba.

        Returns:
            A list of Dispatcher objects. Empty if no dispatchers found.
        """
        # Lazy import to avoid circular dependencies
        from d810.optimizers.microcode.flow.flattening.unflattener import (
            OllvmDispatcherInfo,
        )
        from d810.optimizers.microcode.flow.flattening.dispatcher_detection import (
            DispatcherCache,
        )

        mba = context.mba
        dispatchers: List[Dispatcher] = []

        # Use the existing OllvmDispatcherInfo for analysis
        # This wraps the proven logic while providing a clean interface
        disp_info = OllvmDispatcherInfo(mba)

        # Step 1: Guess the outermost dispatcher block
        outmost_dispatch_num = disp_info.guess_outmost_dispatcher_blk()
        if outmost_dispatch_num == -1:
            context.logger.debug(
                "No dispatcher candidates found (guess_outmost_dispatcher_blk)"
            )
            return dispatchers

        context.logger.debug(
            f"Guessed outermost dispatcher at block {outmost_dispatch_num}"
        )

        # Step 2: Get candidate blocks using DispatcherCache
        if self._use_dispatcher_cache:
            cache = DispatcherCache.get_or_create(mba)
            analysis = cache.analyze()
            # Convert dispatcher serials to blocks
            candidate_blocks = [
                mba.get_mblock(serial) for serial in analysis.dispatchers
                if mba.get_mblock(serial) is not None
            ]
        else:
            # Brute force: check all blocks
            candidate_blocks = [
                mba.get_mblock(i) for i in range(mba.qty)
                if mba.get_mblock(i) is not None
            ]

        # Step 3: Explore each candidate to validate and extract info
        for blk in candidate_blocks:
            if blk is None:
                continue

            # Reset and explore from this block
            disp_info.outmost_dispatch_num = outmost_dispatch_num
            disp_info.last_num_in_first_blks = disp_info.get_last_blk_in_first_blks()

            # Explore validates and populates dispatcher info
            is_valid = disp_info.explore(
                blk,
                min_entropy=self._min_entropy,
                max_entropy=self._max_entropy
            )

            if is_valid:
                # Convert to immutable Dispatcher dataclass
                dispatcher = self._convert_to_dispatcher(disp_info, mba)
                if dispatcher:
                    dispatchers.append(dispatcher)
                    context.logger.info(
                        f"Found dispatcher: entry={dispatcher.entry_block.serial}, "
                        f"internal={len(dispatcher.internal_blocks)}, "
                        f"exits={len(dispatcher.exit_blocks)}"
                    )

        context.logger.info(f"Found {len(dispatchers)} O-LLVM dispatcher(s)")
        return dispatchers

    def _convert_to_dispatcher(
        self,
        disp_info: "OllvmDispatcherInfo",
        mba: ida_hexrays.mba_t
    ) -> Dispatcher | None:
        """Convert mutable OllvmDispatcherInfo to immutable Dispatcher.

        Args:
            disp_info: The dispatcher info from exploration.
            mba: The microcode array.

        Returns:
            An immutable Dispatcher dataclass, or None if conversion fails.
        """
        # Guard against invalid state
        if not disp_info.entry_block:
            return None

        entry_block = disp_info.entry_block.blk

        # Extract the state variable (mop_compared from entry block)
        state_variable = disp_info.mop_compared
        if state_variable is None:
            return None

        # Convert internal blocks from OllvmDispatcherBlockInfo to mblock_t
        internal_blocks = [
            block_info.blk
            for block_info in disp_info.dispatcher_internal_blocks
            if block_info.blk is not None
        ]

        # Convert exit blocks
        exit_blocks = [
            block_info.blk
            for block_info in disp_info.dispatcher_exit_blocks
            if block_info.blk is not None
        ]

        return Dispatcher(
            entry_block=entry_block,
            state_variable=state_variable,
            comparison_values=list(disp_info.comparison_values),
            internal_blocks=internal_blocks,
            exit_blocks=exit_blocks,
            mba=mba
        )

    def find_single(
        self,
        context: OptimizationContext,
        start_block: ida_hexrays.mblock_t
    ) -> Dispatcher | None:
        """Find a single dispatcher starting from a specific block.

        This is useful when you already know which block to analyze,
        avoiding the full scan.

        Args:
            context: The optimization context.
            start_block: The block to analyze as potential dispatcher entry.

        Returns:
            A Dispatcher if found, None otherwise.
        """
        from d810.optimizers.microcode.flow.flattening.unflattener import (
            OllvmDispatcherInfo,
        )

        mba = context.mba
        disp_info = OllvmDispatcherInfo(mba)

        # Setup and explore
        disp_info.outmost_dispatch_num = disp_info.guess_outmost_dispatcher_blk()
        disp_info.last_num_in_first_blks = disp_info.get_last_blk_in_first_blks()

        is_valid = disp_info.explore(
            start_block,
            min_entropy=self._min_entropy,
            max_entropy=self._max_entropy
        )

        if is_valid:
            return self._convert_to_dispatcher(disp_info, mba)

        return None



"""Flow optimizer package.

Import rule modules here so their FlowOptimizationRule subclasses register
eagerly via Registrant metaclass side effects.
"""

try:
    import ida_hexrays  # noqa: F401
except ImportError:
    # Allow package import in non-IDA environments.
    pass
else:
    # Registration side-effects for commonly used flow rules.
    from d810.optimizers.microcode.flow.constant_prop import global_const_inline  # noqa: F401
    from d810.optimizers.microcode.flow.flattening import block_merge  # noqa: F401
    from d810.optimizers.microcode.flow.flattening import mba_state_preconditioner  # noqa: F401
    from d810.optimizers.microcode.flow.jumps import indirect_branch  # noqa: F401
    from d810.optimizers.microcode.flow.jumps import indirect_call  # noqa: F401
    from d810.optimizers.microcode.flow import identity_call  # noqa: F401



import collections
import importlib
import logging
import os
import re
import sys
import time
import weakref

# Added for easier path manipulations
from pathlib import Path
from d810.core.typing import TextIO
from unittest import loader, runner, suite

import ida_kernwin

from d810.core.logging import getLogger

# Import Qt classes through qt_shim for compatibility
# All compatibility shims are automatically set up by qt_shim
from d810.qt_shim import QtCore, QtGui, QtWidgets, get_text_margins_as_tuple

# Configure a logger for the script
LOGGER = getLogger(__name__)


def _is_dark_theme() -> bool:
    """Detect if the current Qt theme uses a dark background."""
    try:
        app = QtWidgets.QApplication.instance()
        if app:
            bg = app.palette().color(app.palette().Window)
            # Luminance formula: dark if < 128
            return (bg.red() * 299 + bg.green() * 587 + bg.blue() * 114) / 1000 < 128
    except Exception:
        # Theme probing is best-effort; fall back to dark palette defaults.
        pass
    return True  # Default to dark (IDA Pro 9 default)


# --- Theme-Adaptive Colors ---
if _is_dark_theme():
    # Dark mode - bright colors
    COLORS = {
        "default": "#CCCCCC",
        "running": "#42A5F5",   # bright cyan
        "pass": "#66BB6A",      # bright green
        "fail": "#EF5350",      # bright red
        "error": "#FF7043",     # orange-red
        "skip": "#9E9E9E",      # gray
        "info_log": "#B0B0B0",
        "success_log": "#66BB6A",
        "error_log": "#EF5350",
        "warning_log": "#FFA726",
    }
else:
    # Light mode - darker colors
    COLORS = {
        "default": "#333333",
        "running": "#1565C0",   # dark blue
        "pass": "#2E7D32",      # dark green
        "fail": "#C62828",      # dark red
        "error": "#BF360C",     # dark orange
        "skip": "#757575",      # dark gray
        "info_log": "#616161",
        "success_log": "#2E7D32",
        "error_log": "#C62828",
        "warning_log": "#E65100",
    }


# --- Test Results ---
TEST_RESULT_NONE = 0
TEST_RESULT_RUNNING = 1
TEST_RESULT_SKIP = 2
TEST_RESULT_PASS = 3
TEST_RESULT_EXPECTED_FAIL = 4
TEST_RESULT_UNEXPECTED_PASS = 5
TEST_RESULT_FAIL = 6
TEST_RESULT_ERROR = 7

# --- Partial Run Modes ---
RUN_TEST_SETUP_ONLY = 0
RUN_TEST_NO_TEAR_DOWN = 1
RUN_TEST_FULL = 2

# --- Tree Item Categories ---
ITEM_CATEGORY_ALL = 0
ITEM_CATEGORY_PACKAGE = 1
ITEM_CATEGORY_MODULE = 2
ITEM_CATEGORY_SUITE = 3
ITEM_CATEGORY_TEST = 4

# --- UI Log Colors (theme-adaptive) ---
LOG_COLOR_INFORMATION = QtGui.QColor(COLORS["info_log"])
LOG_COLOR_ERROR = QtGui.QColor(COLORS["error_log"])
LOG_COLOR_FAILED = QtGui.QColor(COLORS["fail"])
LOG_COLOR_WARNING = QtGui.QColor(COLORS["warning_log"])
LOG_COLOR_SUCCESS = QtGui.QColor(COLORS["success_log"])


def make_main_layout(host_widget=None):
    _lay = QtWidgets.QVBoxLayout(host_widget)
    _lay.setContentsMargins(6, 6, 6, 6)
    _lay.setSpacing(3)
    return _lay


def make_minor_horizontal_layout():
    _layout = QtWidgets.QHBoxLayout()
    _layout.setSpacing(3)
    _layout.setContentsMargins(0, 0, 0, 0)
    return _layout


def make_icon_button(icon=None, parent=None):
    btn = QtWidgets.QPushButton("", parent)
    if icon:
        btn.setIcon(icon)
    btn.setFlat(True)
    btn.setIconSize(QtCore.QSize(20, 20))
    btn.setFixedSize(QtCore.QSize(24, 24))
    return btn


def make_menu_tool_button(icon=None, tool_tip="", parent=None):
    btn = QtWidgets.QToolButton(parent)
    if icon:
        btn.setIcon(icon)
    btn.setToolTip(tool_tip)
    btn.setContentsMargins(0, 0, 0, 0)
    btn.setIconSize(QtCore.QSize(20, 20))
    btn.setPopupMode(btn.InstantPopup)
    menu = QtWidgets.QMenu(btn)
    btn.setMenu(menu)
    return (btn, menu)


def is_path(path):
    return "/" in path or "\\" in path


def object_from_dot_path(dot_path, silent=False):
    if not dot_path:
        return None

    def try_import_closest_module(paths):
        module = None
        while paths:
            try:
                module_name = ".".join(paths)
                # Use importlib for cleaner imports
                if module_name in sys.modules:
                    module = importlib.reload(sys.modules[module_name])
                else:
                    module = importlib.import_module(module_name)
                break
            except ImportError:
                del paths[-1]
                if not paths:
                    return None
        return module

    parts = dot_path.split(".")
    module = try_import_closest_module(parts[:])
    if not module:
        if not silent:
            LOGGER.error("No module found from %s", dot_path)
        return None

    obj = module
    for part in parts[1:]:
        try:
            obj = getattr(obj, part)
        except AttributeError:
            if not silent:
                LOGGER.exception("Error importing the module at path %s", dot_path)
            return None
    return obj


def parse_parameterized_test_id(test_id):
    is_parameterized = ":" in test_id
    return (
        (is_parameterized, test_id.split(":")[0])
        if is_parameterized
        else (is_parameterized, test_id)
    )


class TestRunInfo(object):
    def __init__(self):
        self.reset()

    def reset(self):
        self.run_test_ids = []
        self.run_count = 0
        self.failed_test_id = None
        self.success_count = 0
        self.failed_count = 0
        self.error_count = 0
        self.skip_count = 0
        self.expected_failure_count = 0
        self.unexpected_success_count = 0
        self._session_start_time = 0.0
        self.session_run_time = 0.0
        self._test_start_times = {}
        self.single_test_run_time = 0


class TestManager:
    def __init__(self, ui, start_dir_or_module, top_dir=None):
        self._ui = ui
        self._start_dir_or_module = start_dir_or_module or ""
        self._top_dir = top_dir or ""
        self._stop_on_error = False
        self._runner = PyUnitRunner(self)
        self.set_dirs(start_dir_or_module, top_dir)

    def set_dirs(self, start_dir_or_module, top_dir=None):
        self._start_dir_or_module = start_dir_or_module or ""
        self._top_dir = top_dir or ""
        if self._start_dir_or_module and not self._top_dir:
            if os.path.isdir(self._start_dir_or_module):
                self._top_dir = os.path.dirname(self._start_dir_or_module)

    def start_dir_or_module(self):
        return self._start_dir_or_module

    def top_dir(self):
        return self._top_dir

    def set_stop_on_error(self, stop):
        self._stop_on_error = stop

    def reload_test_modules(self):
        """
        Reload all previously imported test modules under the current start_dir.
        """
        start_dir = self._start_dir_or_module
        top_dir = self._top_dir
        if not start_dir or not top_dir:
            return
        # Ensure tests are importable
        if top_dir not in sys.path:
            sys.path.insert(0, top_dir)
        # Compute package name from start_dir relative to top_dir
        relpath = os.path.relpath(start_dir, top_dir)
        package_name = relpath.replace(os.sep, ".")
        # Reload modules matching the test package
        for module_name, module in list(sys.modules.items()):
            if module_name == package_name or module_name.startswith(
                package_name + "."
            ):
                try:
                    importlib.reload(module)
                    LOGGER.info("Reloaded test module %s", module_name)
                except Exception:
                    LOGGER.exception("Failed to reload test module %s", module_name)

    def stop_on_error(self):
        """
        Return whether to stop on the first error.
        """
        return self._stop_on_error

    def run_tests(self, *tests):
        # Reload test modules so changes are picked up
        self.reload_test_modules()
        self._runner.run_tests(*tests)

    def iter_all_test_ids(self):
        yield from self._runner.iter_all_test_ids()

    def run_all_tests(self):
        tests = list(self.iter_all_test_ids())
        if not tests:
            LOGGER.warning("No tests found to run.")
            return
        self.run_tests(*tests)

    def run_single_test_partially(self, test_id, partial_mode):
        self._runner.run_single_test_partially(test_id, partial_mode)

    def has_last_lister_error(self):
        return self._runner.has_last_lister_error()

    def last_run_test_ids(self):
        return self._runner.last_run_test_ids()

    def last_run_info(self):
        return self._runner.last_run_info()

    def last_failed_test_id(self):
        return self._runner.last_failed_test_id()


class UiStream(TextIO):
    _test_main_window = None
    _log_browser = None
    _url_template = "<a href='?{1}={2}'>{0}</a>"
    _trace_exp = None

    @classmethod
    def set_ui(cls, wgt):
        cls._test_main_window = weakref.ref(wgt)
        cls._log_browser = cls.call_ui_method("get_log_browser_widget")

    @classmethod
    def unset_ui(cls, wgt):
        if cls._test_main_window and cls._test_main_window() == wgt:
            cls._test_main_window = None

    @classmethod
    def log_browser(cls):
        return cls._log_browser

    def __init__(self):
        if not self.__class__._trace_exp:
            self.__class__._trace_exp = re.compile(r'File "(.*?)", line (\d+),')
        self._test_result = None

    def set_result(self, result=None):
        self._test_result = result

    def write(self, msg):
        report_ui = self.log_browser()
        if not report_ui:
            sys.stdout.write(msg)  # Fallback
            return

        if self._test_result == TEST_RESULT_ERROR:
            report_ui.log_error(msg)
        elif self._test_result == TEST_RESULT_FAIL:
            report_ui.log_failed(msg)
        elif self._test_result == TEST_RESULT_SKIP:
            report_ui.log_warning(msg)
        elif self._test_result == TEST_RESULT_PASS:
            report_ui.log_success(msg)
        else:
            report_ui.log_information(msg)

    def writeln(self, msg=None):
        if msg:
            self.write(msg)
        self.write("<br>")

    def flush(self):
        pass

    @classmethod
    def call_ui_method(cls, method_name, *args, **kwargs):
        if not cls._test_main_window:
            return
        ui = cls._test_main_window()
        if ui:
            method = getattr(ui, method_name, None)
            if method:
                return method(*args, **kwargs)


class LogHandler(logging.Handler):
    def emit(self, record):
        report_ui = UiStream.log_browser()
        if not report_ui:
            return
        msg = self.format(record)
        if record.levelno >= logging.ERROR:
            report_ui.log_failed(msg)
        elif record.levelno >= logging.WARNING:
            report_ui.log_warning(msg)
        else:
            report_ui.log_information(msg)


class PyUnitUiMixin(object):
    last_run_info = TestRunInfo()
    _original_std_out = sys.stdout
    _original_std_err = sys.stderr

    def __init__(self, stream: UiStream):
        self.stream: UiStream = stream
        self.log_handler = LogHandler()
        self.std_out_capturer = None  # Will be created on demand
        self.std_err_capturer = None  # Will be created on demand

    @classmethod
    def reset_last_data(cls):
        cls.last_run_info.reset()

    def _call_ui_method(self, method, *args, **kwargs):
        UiStream.call_ui_method(method, *args, **kwargs)

    @classmethod
    def _record_last_failed_test_id(cls, test_id):
        if not cls.last_run_info.failed_test_id:
            cls.last_run_info.failed_test_id = test_id

    def _at_outcome_available(self, test_id, result_code):
        cls = type(self)
        if result_code == TEST_RESULT_ERROR:
            cls.last_run_info.error_count += 1
            self._record_last_failed_test_id(test_id)
        elif result_code == TEST_RESULT_FAIL:
            cls.last_run_info.failed_count += 1
            self._record_last_failed_test_id(test_id)
        elif result_code == TEST_RESULT_EXPECTED_FAIL:
            cls.last_run_info.expected_failure_count += 1
        elif result_code == TEST_RESULT_UNEXPECTED_PASS:
            cls.last_run_info.unexpected_success_count += 1
        elif result_code == TEST_RESULT_SKIP:
            cls.last_run_info.skip_count += 1
        elif result_code == TEST_RESULT_PASS:
            cls.last_run_info.success_count += 1

        self._call_ui_method("show_result_on_item_by_test_id", test_id, result_code)
        self._stop_log_processors()

    def _at_start_test_run(self):
        self._call_ui_method("repaint_ui")
        cls = type(self)
        cls.last_run_info.reset()
        cls.last_run_info._session_start_time = time.time()
        self._call_ui_method("on_test_running_session_start")

    def _start_log_processors(self):
        logging.getLogger().addHandler(self.log_handler)
        sys.stdout = self.stream
        sys.stderr = self.stream

    def _stop_log_processors(self):
        logging.getLogger().removeHandler(self.log_handler)
        sys.stdout = self._original_std_out
        sys.stderr = self._original_std_err

        # Some third-party code (or the standard library itself when
        # `logging.basicConfig(stream=sys.stdout...` was invoked while the
        # tests were running) may have created new `logging.StreamHandler`
        # instances bound to the *current* value of ``sys.stdout`` – i.e. the
        # temporary `UiStream` (which is what we use to display the logs for the test).
        #
        # When we restore the original `sys.stdout`/`sys.stderr` above, those
        # handlers still hold a stale reference to the deleted `LogBrowser`
        # which ultimately crashes when the UI has been closed.
        #
        # To fix this, we walk *all* known loggers and patch any
        # `StreamHandler` whose ``stream`` attribute is that stale object so
        # that it points back to the real console stream.
        def _all_loggers():
            """Yield the root logger and every registered named logger."""
            yield logging.getLogger()
            for lg in logging.Logger.manager.loggerDict.values():
                # Some entries in loggerDict can be placeholders (Loggers vs dict)
                if isinstance(lg, logging.Logger):
                    yield lg

        for lg in _all_loggers():
            for hdlr in list(lg.handlers):
                if (
                    isinstance(hdlr, logging.StreamHandler)
                    and hdlr.stream is self.stream
                ):
                    hdlr.stream = self._original_std_out

    def _at_start_test(self, test):
        cls = type(self)
        cls.last_run_info.run_count += 1
        original_test_id = test.id()
        _, test_id = parse_parameterized_test_id(original_test_id)
        if test_id not in cls.last_run_info.run_test_ids:
            test_start_time = time.time()
            cls.last_run_info._test_start_times[test_id] = test_start_time
            cls.last_run_info.run_test_ids.append(test_id)
            self._call_ui_method("on_single_test_start", test_id, test_start_time)

    def _at_stop_test(self, test):
        original_test_id = test.id()
        _, test_id = parse_parameterized_test_id(original_test_id)
        stop_time = time.time()
        cls = type(self)
        test_start_time = cls.last_run_info._test_start_times.get(
            test_id, cls.last_run_info._session_start_time
        )
        cls.last_run_info.single_test_run_time = stop_time - test_start_time
        self._call_ui_method("on_single_test_stop", test_id, stop_time)
        self._call_ui_method("repaint_ui")

    def _at_stop_test_run(self):
        cls = type(self)
        cls.last_run_info.session_run_time = (
            time.time() - cls.last_run_info._session_start_time
        )
        self._call_ui_method("on_all_tests_finished")
        # Ensure that standard output/error streams and the root logger are
        # restored to their original state when the complete test run
        # finishes.  Without this call, further logging performed outside the
        # testbed session would continue to be redirected to the UI log
        # browser instead of the console.
        self._stop_log_processors()


class PyUnitTestRunnerWrapper(runner.TextTestRunner):
    def __init__(self, verbosity=2, failfast=False, partial_mode=RUN_TEST_FULL):
        self.stream = UiStream()
        self._partial_mode = partial_mode

        # Call the parent constructor without explicitly passing our custom
        # ``resultclass``.  This avoids mypy’s type-checking complaint that
        # ``PyUnitTestResult`` expects a ``UiStream`` rather than a ``TextIO``.
        super().__init__(
            stream=self.stream,
            verbosity=verbosity,
            failfast=failfast,
        )

        # Now that the base class is initialized we can safely set the
        # desired result class.  The ``type: ignore`` suppresses the
        # mismatch warning between ``UiStream`` and ``TextIO`` in stubs.
        self.resultclass = PyUnitTestResult  # type: ignore[assignment]

    @staticmethod
    def _dummy_function(*_, **__):
        pass

    def run(self, test):
        if self._partial_mode != RUN_TEST_FULL:
            test_case = test
            while isinstance(test_case, suite.TestSuite):
                test_case = next(iter(test_case), None)

            if test_case:
                print(type(test_case))
                if hasattr(test_case, "tearDown"):
                    test_case.tearDown = self._dummy_function
                if self._partial_mode == RUN_TEST_SETUP_ONLY:
                    method_name = test_case.id().split(".")[-1]
                    if hasattr(test_case, method_name):
                        setattr(test_case, method_name, self._dummy_function)
                    LOGGER.info("Run %s.setUp() only:", test_case.__class__.__name__)
                elif self._partial_mode == RUN_TEST_NO_TEAR_DOWN:
                    LOGGER.info("Skipped %s.tearDown():", test_case.__class__.__name__)
        return super().run(test)


class PyUnitTestResult(runner.TextTestResult, PyUnitUiMixin):

    def __init__(self, stream: UiStream, descriptions: bool, verbosity: int):
        self.Cls = self.__class__
        super().__init__(stream, descriptions, verbosity)
        self.stream: UiStream = stream
        PyUnitUiMixin.__init__(self, stream)

    def startTestRun(self):
        self._at_start_test_run()
        super().startTestRun()

    def startTest(self, test):
        self._at_start_test(test)
        super().startTest(test)
        self._start_log_processors()

    def stopTest(self, test):
        super().stopTest(test)
        self._at_stop_test(test)

    def stopTestRun(self):
        super().stopTestRun()
        self._at_stop_test_run()
        self.stream.set_result(
            TEST_RESULT_PASS if self.wasSuccessful() else TEST_RESULT_ERROR
        )

    def addSuccess(self, test):
        self.stream.set_result(TEST_RESULT_PASS)
        super().addSuccess(test)
        self._at_outcome_available(test.id(), TEST_RESULT_PASS)

    def addError(self, test, err):
        self.stream.set_result(TEST_RESULT_ERROR)
        super().addError(test, err)
        self._at_outcome_available(test.id(), TEST_RESULT_ERROR)

    def addFailure(self, test, err):
        self.stream.set_result(TEST_RESULT_FAIL)
        super().addFailure(test, err)
        self._at_outcome_available(test.id(), TEST_RESULT_FAIL)

    def addSkip(self, test, reason):
        self.stream.set_result(TEST_RESULT_SKIP)
        super().addSkip(test, reason)
        self._at_outcome_available(test.id(), TEST_RESULT_SKIP)

    def addExpectedFailure(self, test, err):
        self.stream.set_result(TEST_RESULT_EXPECTED_FAIL)
        super().addExpectedFailure(test, err)
        self._at_outcome_available(test.id(), TEST_RESULT_EXPECTED_FAIL)

    def addUnexpectedSuccess(self, test):
        self.stream.set_result(TEST_RESULT_UNEXPECTED_PASS)
        super().addUnexpectedSuccess(test)
        self._at_outcome_available(test.id(), TEST_RESULT_UNEXPECTED_PASS)


class BaseTestRunner:
    def __init__(self, manager: TestManager):
        self._manager: TestManager = manager

    def run_tests(self, *test_ids):
        raise NotImplementedError

    def iter_all_test_ids(self):
        raise NotImplementedError

    def run_single_test_partially(self, test_id, partial_mode):
        raise NotImplementedError

    @classmethod
    def last_run_info(cls):
        return TestRunInfo()

    @classmethod
    def last_run_test_ids(cls):
        return cls.last_run_info().run_test_ids

    @classmethod
    def last_failed_test_id(cls):
        return cls.last_run_info().failed_test_id

    @classmethod
    def has_last_lister_error(cls):
        return False


class PyUnitRunner(BaseTestRunner):
    _got_error = False

    def run_tests(self, *test_ids):
        self._run_tests(RUN_TEST_FULL, *test_ids)

    def _run_tests(self, partial_mode=RUN_TEST_FULL, *test_ids):
        failfast = self._manager.stop_on_error()
        PyUnitTestResult.reset_last_data()
        test_runner = PyUnitTestRunnerWrapper(
            failfast=failfast, partial_mode=partial_mode, verbosity=2
        )
        # Create a suite from the test IDs
        suite = loader.defaultTestLoader.loadTestsFromNames(test_ids)
        test_runner.run(suite)

    def run_single_test_partially(self, test_id, partial_mode):
        self._run_tests(partial_mode, test_id)

    def _collect_all_paths(self, tests):
        if isinstance(tests, suite.TestSuite):
            for t in tests:
                yield from self._collect_all_paths(t)
        else:
            yield tests.id()

    def iter_all_test_ids(self):
        self.__class__._got_error = False
        start_dir = self._manager.start_dir_or_module()
        top_dir = self._manager.top_dir()
        if not start_dir or not os.path.isdir(start_dir):
            LOGGER.warning("Start directory is not a valid directory: %s", start_dir)
            return iter([])

        if not top_dir or not os.path.isdir(top_dir):
            top_dir = os.path.dirname(start_dir)
            LOGGER.info("Top level directory not set, defaulting to: %s", top_dir)

        try:
            # Add top_dir to sys.path to ensure imports work correctly
            if top_dir not in sys.path:
                sys.path.insert(0, top_dir)
            tests = loader.defaultTestLoader.discover(start_dir, top_level_dir=top_dir)
            yield from self._collect_all_paths(tests)
        except Exception:
            self.__class__._got_error = True
            LOGGER.exception("Unable to load tests from %s", start_dir)

    @classmethod
    def has_last_lister_error(cls):
        return cls._got_error

    @classmethod
    def last_run_info(cls):
        return PyUnitTestResult.last_run_info


class InlineButtonLineEdit(QtWidgets.QLineEdit):
    def __init__(self, with_clear_button=False, parent=None):
        super().__init__(parent)
        self._buttons = collections.OrderedDict()
        # Use compatibility helper function for textMargins()
        self._init_margins = get_text_margins_as_tuple(self)
        if with_clear_button:
            self.add_clear_button()

    def add_button(self, btn_id, button):
        button.setFixedSize(QtCore.QSize(20, 20))
        button.setParent(self)
        button.setStyleSheet("QAbstractButton{background:transparent; border: none;}")
        self._buttons[btn_id] = button
        self._refresh_buttons()
        return button

    def add_clear_button(self):
        btn = make_icon_button(
            self.style().standardIcon(QtWidgets.QStyle.SP_LineEditClearButton), self
        )
        btn.setToolTip("Clear")
        btn.setVisible(False)
        btn.clicked.connect(self.clear)
        self.textChanged.connect(lambda txt: btn.setVisible(bool(txt)))
        return self.add_button("__clear__", btn)

    def resizeEvent(self, event):
        super().resizeEvent(event)
        self._refresh_buttons()

    def _refresh_buttons(self):
        # Use initial text margins captured at creation
        init_left, init_top, init_right, init_bottom = self._init_margins
        right_margin = 0
        for btn in reversed(list(self._buttons.values())):
            if btn.isVisible():
                btn.move(
                    self.width() - right_margin - btn.width() - 2,
                    (self.height() - btn.height()) // 2,
                )
                right_margin += btn.width() + 2
        total_right = init_right + right_margin
        self.setTextMargins(init_left, init_top, total_right, init_bottom)


class RootPathEdit(InlineButtonLineEdit):
    root_path_changed = QtCore.pyqtSignal(str, str)

    def __init__(self, parent=None):
        super().__init__(with_clear_button=False, parent=parent)
        self.editingFinished.connect(self.on_edit_finished)
        self._init_path = None

    def set_initial_path(self, path):
        self._init_path = path

    def setText(self, txt):
        super().setText(txt)
        self.set_initial_path(txt)

    def on_edit_finished(self):
        txt = str(self.text())
        if txt != self._init_path:
            self.set_initial_path(txt)
            self.root_path_changed.emit(txt, "")

    def _on_browse_tests_root_dir(self):
        dir_path = QtWidgets.QFileDialog.getExistingDirectory(
            self, "Pick the Test Root Directory"
        )
        if dir_path:
            self.setText(dir_path)
            self.on_edit_finished()


class LogBrowser(QtWidgets.QTextBrowser):
    def __init__(self, parent=None):
        super().__init__(parent)
        fn = self.font()
        fn.setFamily("Courier New")
        fn.setPointSize(10)
        self.setFont(fn)
        self.setReadOnly(True)
        self.setOpenLinks(False)

    def log_with_color(self, msg, color, *args):
        self.moveCursor(QtGui.QTextCursor.End)
        self.setTextColor(color)
        msg_formatted = msg if not args else (msg % args)
        self.insertPlainText(msg_formatted)
        self.moveCursor(QtGui.QTextCursor.End)

    def log_separator(self):
        self.log_warning("\n" + ">" * 70 + "\n")

    def log_information(self, msg, *args):
        self.log_with_color(msg, LOG_COLOR_INFORMATION, *args)

    def log_success(self, msg, *args):
        self.log_with_color(msg, LOG_COLOR_SUCCESS, *args)

    def log_failed(self, msg, *args):
        self.log_with_color(msg, LOG_COLOR_FAILED, *args)

    def log_error(self, msg, *args):
        self.log_with_color(msg, LOG_COLOR_ERROR, *args)

    def log_warning(self, msg, *args):
        self.log_with_color(msg, LOG_COLOR_WARNING, *args)


class StatusLabel(QtWidgets.QLabel):
    def __init__(self, parent):
        super().__init__(parent)
        self.setTextInteractionFlags(QtCore.Qt.TextSelectableByMouse)
        self._test_manager: TestManager

    def set_test_manager(self, manager):
        self._test_manager = manager

    def update_report(self):
        if not self._test_manager:
            return
        run_info = self._test_manager.last_run_info()
        test_lbl = "tests" if run_info.run_count != 1 else "test"
        msgs = [
            f"{run_info.run_count} {test_lbl} run in {run_info.session_run_time:.3f} sec"
        ]
        if run_info.failed_count:
            msgs.append(
                f'<font color="{COLORS["fail"]}">{run_info.failed_count} failed</font>'
            )
        if run_info.error_count:
            msgs.append(
                f'<font color="{COLORS["error"]}">{run_info.error_count} errors</font>'
            )
        if run_info.skip_count:
            msgs.append(
                f'<font color="{COLORS["warning_log"]}">{run_info.skip_count} skipped</font>'
            )
        self.setText(", ".join(msgs))

    def report_test_count(self, test_count):
        self.setText(f"{test_count} tests found.")

    def start_collecting_tests(self):
        self.setText("Loading tests...")
        self.repaint()


class UnitTestTreeView(QtWidgets.QTreeWidget):
    run_all_tests = QtCore.pyqtSignal()
    run_tests = QtCore.pyqtSignal(tuple)
    run_setup_only = QtCore.pyqtSignal(str)
    run_without_tear_down = QtCore.pyqtSignal(str)

    def __init__(self, parent):
        super().__init__(parent)
        self.setColumnCount(2)
        self.setHeaderLabels(["Test", "Time (s)"])
        self.header().setSectionResizeMode(0, QtWidgets.QHeaderView.Stretch)
        self.header().setSectionResizeMode(1, QtWidgets.QHeaderView.ResizeToContents)
        self.setAlternatingRowColors(True)
        self.setExpandsOnDoubleClick(False)
        self.setSelectionMode(self.ExtendedSelection)
        self.itemDoubleClicked.connect(
            lambda item, col: self.on_item_double_clicked(item)
        )
        self.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
        self.customContextMenuRequested.connect(self._make_context_menu)

        self._test_manager: TestManager
        self._all_items_id_map = {}
        self._test_cases = []
        self._root_test_item: QtWidgets.QTreeWidgetItem
        self._filter_text = ""

    def set_test_manager(self, manager):
        self._test_manager = manager

    def on_item_double_clicked(self, item):
        if item is self._root_test_item:
            self.run_all_tests.emit()
        else:
            self.run_tests.emit((self.test_id_of_item(item),))

    def reload(self):
        self.clear()
        self._all_items_id_map.clear()
        self._test_cases.clear()

        start_dir = self._test_manager.start_dir_or_module()
        if not start_dir:
            self._root_test_item = QtWidgets.QTreeWidgetItem(
                self, ["No directory selected."]
            )
            return 0

        # Show only last 2 path components in the tree, full path as tooltip
        path_obj = Path(start_dir)
        display_name = "/".join(path_obj.parts[-2:]) if len(path_obj.parts) >= 2 else start_dir
        self._root_test_item = QtWidgets.QTreeWidgetItem(self, [display_name])
        self._root_test_item.setToolTip(0, start_dir)  # Full path as tooltip
        self._root_test_item.setData(
            0, QtCore.Qt.UserRole, (ITEM_CATEGORY_ALL, start_dir)
        )
        self._set_item_icon_state(self._root_test_item, TEST_RESULT_NONE)
        self.addTopLevelItem(self._root_test_item)
        self._all_items_id_map[start_dir] = self._root_test_item

        test_count = 0
        for test_id in self._test_manager.iter_all_test_ids():
            # Skip unittest discovery failures
            if "_FailedTest" in test_id:
                continue
            test_count += 1
            self._add_test_to_tree(test_id)

        self.expandAll()
        self.header().resizeSections(QtWidgets.QHeaderView.ResizeToContents)
        self.header().setSectionResizeMode(0, QtWidgets.QHeaderView.Stretch)
        return test_count

    def _add_test_to_tree(self, test_id):
        parts = test_id.split(".")
        current_parent = self._root_test_item
        path_so_far = []
        # Build or reuse each segment node
        for part in parts[:-1]:
            path_so_far.append(part)
            new_path = ".".join(path_so_far)
            if new_path in self._all_items_id_map:
                current_parent = self._all_items_id_map[new_path]
            else:
                child = QtWidgets.QTreeWidgetItem(current_parent, [part])
                child.setData(0, QtCore.Qt.UserRole, (ITEM_CATEGORY_MODULE, new_path))
                self._set_item_icon_state(child, TEST_RESULT_NONE)
                self._all_items_id_map[new_path] = child
                current_parent = child
        # Add the actual test method as a leaf
        method_name = parts[-1]
        test_item = QtWidgets.QTreeWidgetItem(current_parent, [method_name])
        test_item.setData(0, QtCore.Qt.UserRole, (ITEM_CATEGORY_TEST, test_id))
        self._set_item_icon_state(test_item, TEST_RESULT_NONE)
        self._all_items_id_map[test_id] = test_item
        self._test_cases.append(test_item)

    def _set_item_icon_state(self, item, state):
        # Theme-adaptive colors
        colors = {
            TEST_RESULT_NONE: QtGui.QColor(COLORS["default"]),
            TEST_RESULT_RUNNING: QtGui.QColor(COLORS["running"]),
            TEST_RESULT_PASS: QtGui.QColor(COLORS["pass"]),
            TEST_RESULT_FAIL: QtGui.QColor(COLORS["fail"]),
            TEST_RESULT_ERROR: QtGui.QColor(COLORS["error"]),
            TEST_RESULT_SKIP: QtGui.QColor(COLORS["skip"]),
        }
        color = colors.get(state, QtGui.QColor(COLORS["default"]))
        item.setForeground(0, QtGui.QBrush(color))
        item.setData(0, QtCore.Qt.UserRole + 1, state)

    def _update_ancestors_state(self, item):
        parent = item.parent()
        if not parent:
            return

        max_state = TEST_RESULT_NONE
        for i in range(parent.childCount()):
            child_state = parent.child(i).data(0, QtCore.Qt.UserRole + 1)
            if child_state is not None and child_state > max_state:
                max_state = child_state

        self._set_item_icon_state(parent, max_state)
        self._update_ancestors_state(parent)

    def test_id_of_item(self, item):
        data = item.data(0, QtCore.Qt.UserRole)
        return data[1] if data and len(data) > 1 else None

    def on_single_test_start(self, test_id, start_time):
        item = self._all_items_id_map.get(test_id)
        if item:
            self._set_item_icon_state(item, TEST_RESULT_RUNNING)
            item.setText(1, "running...")
            self._update_ancestors_state(item)

    def on_single_test_stop(self, test_id, end_time):
        item = self._all_items_id_map.get(test_id)
        if item:
            run_info = self._test_manager.last_run_info()
            item.setText(1, f"{run_info.single_test_run_time:.3f}")

    def show_result_on_item_by_test_id(self, test_id, state):
        item = self._all_items_id_map.get(test_id)
        if item:
            self._set_item_icon_state(item, state)
            self._update_ancestors_state(item)

    def on_all_tests_finished(self):
        pass

    def apply_filter(self, text: str) -> None:
        """Filter tree items based on search text."""
        self._filter_text = text.lower()
        if not self._filter_text:
            self._set_all_visible(self._root_test_item, True)
            return
        self._set_all_visible(self._root_test_item, False)
        self._show_matching(self._root_test_item, self._filter_text)

    def _set_all_visible(self, parent, visible):
        """Recursively set visibility of all items."""
        for i in range(parent.childCount()):
            child = parent.child(i)
            child.setHidden(not visible)
            self._set_all_visible(child, visible)

    def _show_matching(self, parent, text):
        """Recursively show matching items and their ancestors."""
        any_visible = False
        for i in range(parent.childCount()):
            child = parent.child(i)
            child_matches = text in child.text(0).lower()
            descendant_matches = self._show_matching(child, text)
            if child_matches or descendant_matches:
                child.setHidden(False)
                any_visible = True
        return any_visible

    def _make_context_menu(self, pos):
        item = self.itemAt(pos)
        if not item:
            return

        menu = QtWidgets.QMenu()
        test_id = self.test_id_of_item(item)
        if not test_id:
            return

        run_action = menu.addAction("Run Test(s)")
        run_action.triggered.connect(lambda: self.run_tests.emit((test_id,)))

        menu.addSeparator()

        setup_action = menu.addAction("Run setUp() only")
        setup_action.triggered.connect(lambda: self.run_setup_only.emit(test_id))

        no_teardown_action = menu.addAction("Run without tearDown()")
        no_teardown_action.triggered.connect(
            lambda: self.run_without_tear_down.emit(test_id)
        )

        is_test_case = item.data(0, QtCore.Qt.UserRole)[0] == ITEM_CATEGORY_TEST
        setup_action.setEnabled(is_test_case)
        no_teardown_action.setEnabled(is_test_case)

        menu.exec_(self.viewport().mapToGlobal(pos))


# =============================================================================
# MAIN APPLICATION FORM
# =============================================================================


class TestRunnerForm(ida_kernwin.PluginForm):
    def __init__(self):
        super().__init__()
        self._test_manager: TestManager
        self._view: UnitTestTreeView
        self._log_browser: LogBrowser
        self._status_lbl: StatusLabel
        self._run_all_btn: QtWidgets.QPushButton
        self._run_selected_btn: QtWidgets.QPushButton
        self._stop_on_error_cb: QtWidgets.QCheckBox
        self._root_dir_le: RootPathEdit

    def OnCreate(self, ctx):
        self.parent_widget: QtWidgets.QWidget = self.FormToPyQtWidget(ctx)
        self.parent_widget.setWindowTitle("IDA System Tests Runner")
        self.parent_widget.resize(900, 700)

        # Automatically detect the repository-level "tests" directory.
        # We walk up from this file until we find a sibling directory named
        # "tests" and use that as the default root for unit discovery.
        default_root: str | None = None
        try:
            current_path = Path(__file__).resolve()
            for parent in current_path.parents:
                tests_dir = parent / "tests/system"
                if tests_dir.is_dir():
                    default_root = str(tests_dir)
                    break

            if default_root is None:
                LOGGER.warning(
                    "Could not locate a top-level 'tests' directory."
                    " Please select it manually."
                )
        except Exception:
            # Unexpected failure - log but continue with no default.
            LOGGER.exception("Error while searching for default tests directory")

        self._test_manager = TestManager(
            self, start_dir_or_module=default_root, top_dir=None
        )

        main_lay = make_main_layout(self.parent_widget)

        dir_layout = make_minor_horizontal_layout()
        label = QtWidgets.QLabel("Test Root:")
        dir_layout.addWidget(label)
        # add a little extra space before the separator
        dir_layout.addSpacing(6)
        # Add a vertical separator for visual spacing
        separator = QtWidgets.QFrame(self.parent_widget)
        separator.setFrameShape(QtWidgets.QFrame.VLine)
        separator.setFrameShadow(QtWidgets.QFrame.Sunken)
        dir_layout.addWidget(separator)
        # Increase spacing between elements
        dir_layout.setSpacing(8)
        self._root_dir_le = RootPathEdit(self.parent_widget)
        self._root_dir_le.root_path_changed.connect(self._on_test_root_changed)
        # Ensure the path field expands
        self._root_dir_le.setSizePolicy(
            QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed
        )
        dir_layout.addWidget(self._root_dir_le, 1)
        # External browse button next to the line edit
        browse_btn = QtWidgets.QPushButton(self.parent_widget)
        browse_btn.setIcon(
            self.parent_widget.style().standardIcon(QtWidgets.QStyle.SP_DirOpenIcon)
        )
        browse_btn.setToolTip("Browse")
        browse_btn.clicked.connect(self._root_dir_le._on_browse_tests_root_dir)
        dir_layout.addWidget(browse_btn)
        self._stop_on_error_cb = QtWidgets.QCheckBox("Stop on Error")
        self._stop_on_error_cb.toggled.connect(self._test_manager.set_stop_on_error)
        dir_layout.addWidget(self._stop_on_error_cb)
        main_lay.addLayout(dir_layout)

        # Add filter bar
        filter_layout = make_minor_horizontal_layout()
        filter_label = QtWidgets.QLabel("Filter:")
        filter_layout.addWidget(filter_label)
        self._filter_edit = QtWidgets.QLineEdit(self.parent_widget)
        self._filter_edit.setPlaceholderText("Filter tests...")
        self._filter_edit.setClearButtonEnabled(True)
        filter_layout.addWidget(self._filter_edit, 1)
        main_lay.addLayout(filter_layout)

        splitter = QtWidgets.QSplitter(QtCore.Qt.Vertical, self.parent_widget)
        main_lay.addWidget(splitter, 1)

        tree_widget = QtWidgets.QWidget()
        tree_layout = QtWidgets.QVBoxLayout(tree_widget)
        tree_layout.setContentsMargins(0, 0, 0, 0)
        self._view = UnitTestTreeView(tree_widget)
        self._view.set_test_manager(self._test_manager)
        tree_layout.addWidget(self._view)
        splitter.addWidget(tree_widget)

        log_widget = QtWidgets.QWidget()
        log_layout = QtWidgets.QVBoxLayout(log_widget)
        log_layout.setContentsMargins(0, 0, 0, 0)
        log_layout.addWidget(QtWidgets.QLabel("Log Output"))
        self._log_browser = LogBrowser(log_widget)
        log_layout.addWidget(self._log_browser)
        splitter.addWidget(log_widget)

        splitter.setSizes([400, 300])

        self._status_lbl = StatusLabel(self.parent_widget)
        self._status_lbl.set_test_manager(self._test_manager)
        self._status_lbl.setContentsMargins(0, 2, 0, 2)  # Compact status bar
        main_lay.addWidget(self._status_lbl)

        btn_layout = make_minor_horizontal_layout()
        # Add buttons: Reload Tests, Run Selected, Run All
        btn_layout.addStretch(1)
        self._reload_btn = QtWidgets.QPushButton("Reload Tests")
        self._reload_btn.clicked.connect(self.reload_tests)
        btn_layout.addWidget(self._reload_btn)
        self._run_selected_btn = QtWidgets.QPushButton("Run Selected")
        self._run_selected_btn.clicked.connect(self._run_view_selected_tests)
        btn_layout.addWidget(self._run_selected_btn)
        self._run_all_btn = QtWidgets.QPushButton("Run All")
        self._run_all_btn.clicked.connect(self._run_all_tests)
        btn_layout.addWidget(self._run_all_btn)
        main_lay.addLayout(btn_layout)

        self._view.run_all_tests.connect(self._run_all_tests)
        self._view.run_tests.connect(self._run_tests)
        self._view.run_setup_only.connect(self._run_test_setup_only)
        self._view.run_without_tear_down.connect(self._run_test_without_tear_down)

        # Connect filter text changed signal
        self._filter_edit.textChanged.connect(self._view.apply_filter)

        UiStream.set_ui(self)

        # Prepopulate the root path selector and load tests
        try:
            default_root = self._test_manager.start_dir_or_module()
            if default_root:
                self._root_dir_le.setText(default_root)
                self._root_dir_le.set_initial_path(default_root)
                self.reload_tests()
        except Exception:
            LOGGER.exception("Failed to set initial test view")

    def OnClose(self, form):
        UiStream.unset_ui(self)

    def _on_test_root_changed(self, start_dir, top_dir):
        self._test_manager.set_dirs(start_dir, top_dir)
        self.reload_tests()

    def reload_tests(self):
        self._status_lbl.start_collecting_tests()
        self._test_manager.reload_test_modules()
        test_count = self._view.reload()
        self.repaint_ui()
        self._status_lbl.report_test_count(test_count)

    def _run_all_tests(self):
        self._before_running_tests()
        self._test_manager.run_all_tests()

    def _run_tests(self, test_ids):
        if not test_ids:
            return
        self._before_running_tests()
        self._test_manager.run_tests(*test_ids)

    def _run_view_selected_tests(self):
        selected_ids = [
            self._view.test_id_of_item(item) for item in self._view.selectedItems()
        ]
        if selected_ids:
            self._run_tests(tuple(filter(None, selected_ids)))

    def _run_test_setup_only(self, test_id):
        self._before_running_tests()
        self._test_manager.run_single_test_partially(test_id, RUN_TEST_SETUP_ONLY)

    def _run_test_without_tear_down(self, test_id):
        self._before_running_tests()
        self._test_manager.run_single_test_partially(test_id, RUN_TEST_NO_TEAR_DOWN)

    def _before_running_tests(self):
        self._log_browser.clear()
        self._log_browser.log_separator()

    def get_log_browser_widget(self):
        return self._log_browser

    def on_test_running_session_start(self):
        self._status_lbl.setText("Running tests...")

    def on_all_tests_finished(self):
        self._status_lbl.update_report()
        self._view.on_all_tests_finished()

    def on_single_test_start(self, test_id, start_time):
        self._view.on_single_test_start(test_id, start_time)

    def on_single_test_stop(self, test_id, end_time):
        self._view.on_single_test_stop(test_id, end_time)
        self._status_lbl.update_report()

    def show_result_on_item_by_test_id(self, test_id, state):
        self._view.show_result_on_item_by_test_id(test_id, state)

    def repaint_ui(self):
        QtWidgets.QApplication.processEvents(QtCore.QEventLoop.ExcludeUserInputEvents)


# =============================================================================
# CONVENIENCE HELPER
# =============================================================================


def show_gui():
    """Launch or focus the IDA test runner GUI."""
    form_title = "IDA Unit Test Runner"
    form = ida_kernwin.find_widget(form_title)
    if form:
        ida_kernwin.activate_widget(form, True)
    else:
        form = TestRunnerForm()
        form.Show(
            form_title,
            options=ida_kernwin.PluginForm.WOPN_TAB
            | ida_kernwin.PluginForm.WCLS_CLOSE_LATER,
        )


# Example of how to run this in IDA
if __name__ == "__main__":
    show_gui()



"""Base pattern class for ctree pattern matching.

All ctree patterns inherit from :class:`BasePat`.  The ``check()`` method
is the main entry point; subclasses override it via the ``@base_check``
decorator which handles ``check_op`` validation, ``bind_name`` storage,
and debug output.

Ported from herast (herast/tree/patterns/base_pattern.py).
"""
from __future__ import annotations

import traceback
from d810.core import typing
from d810.core import getLogger
from d810.ctree.match_context import MatchContext

logger = getLogger("D810.ctree")


class BasePat:
    """Base class for all ctree patterns."""

    op: int | None = None

    def __init__(
        self,
        bind_name: str | None = None,
        debug: bool = False,
        debug_msg: str | None = None,
        debug_trace_depth: int = 0,
        check_op: int | None = None,
    ) -> None:
        """
        :param bind_name: should successfully matched item be remembered
        :param debug: should provide debug information during matching
        :param debug_msg: additional message to print on debug
        :param debug_trace_depth: additional trace information on debug
        :param check_op: what item type to check. skips this check if None
        """
        self.bind_name = bind_name
        self.check_op = check_op
        self.debug = debug
        self.debug_msg = debug_msg
        self.debug_trace_depth = debug_trace_depth

    def check(self, item: typing.Any, ctx: MatchContext) -> bool:
        """Base matching operation.

        :param item: AST item (citem_t)
        :param ctx: matching context
        """
        raise NotImplementedError("This is an abstract class")

    @classmethod
    def get_opname(cls) -> str | None:
        """Return the human-readable name of this pattern's op code."""
        from d810.ctree import consts as consts_mod
        return consts_mod.op2str.get(cls.op, None)

    @staticmethod
    def base_check(func: typing.Callable) -> typing.Callable:
        """Decorator for child classes instead of inheritance, since
        before and after calls are needed.
        """

        def __perform_base_check(self: BasePat, item: typing.Any, ctx: MatchContext) -> bool:
            if item is None:
                return False

            if self.check_op is not None and item.op != self.check_op:
                return False

            rv: bool = func(self, item, ctx)

            if rv and self.bind_name is not None:
                rv = ctx.bind_item(self.bind_name, item)

            if self.debug:
                if self.debug_msg:
                    logger.debug("Debug: value = %s, %s", rv, self.debug_msg)
                else:
                    logger.debug("Debug: value = %s", rv)

                if self.debug_trace_depth != 0:
                    logger.debug(
                        "Debug calltrace, address of item: %#x (%s)",
                        item.ea,
                        item.opname,
                    )
                    logger.debug("---------------------------------")
                    for line in traceback.format_stack()[: self.debug_trace_depth]:
                        logger.debug(line.rstrip())
                    logger.debug("---------------------------------")
            return rv

        return __perform_base_check

    @property
    def children(self) -> tuple:
        """Return child patterns. Must be overridden by subclasses."""
        raise NotImplementedError("An abstract class doesn't have any children")



from d810.core import typing

import ida_hexrays

from d810.core import typing
from d810.core import getLogger
from d810.errors import AstEvaluationException
from d810.expr.ast import AstConstant, AstNode, AstProxy, minsn_to_ast
from d810.expr.z3_utils import z3_check_mop_equality
from d810.hexrays.hexrays_formatters import format_minsn_t
from d810.optimizers.microcode.instructions.z3.handler import Z3Rule

logger = getLogger(__name__)


class Z3ConstantOptimization(Z3Rule):
    DESCRIPTION = "Detect and replace obfuscated constants"

    def __init__(self):
        super().__init__()
        self.min_nb_opcode = 3
        self.min_nb_constant = 3

    @property
    def PATTERN(self) -> AstNode | None:
        """Return the pattern to match."""
        return

    @property
    def REPLACEMENT_PATTERN(self) -> AstNode:
        return AstNode(ida_hexrays.m_mov, AstConstant("c_res"))

    @typing.override
    def configure(self, kwargs):
        super().configure(kwargs)
        if "min_nb_opcode" in kwargs.keys():
            self.min_nb_opcode = kwargs["min_nb_opcode"]
        if "min_nb_constant" in kwargs.keys():
            self.min_nb_constant = kwargs["min_nb_constant"]

    @typing.override
    def check_and_replace(self, blk: ida_hexrays.mblock_t, instruction: ida_hexrays.minsn_t) -> ida_hexrays.minsn_t | None:
        tmp = minsn_to_ast(instruction)
        if tmp is None:
            return None
        leaf_info_list, cst_leaf_values, opcodes = tmp.get_information()
        leaf_num = len(leaf_info_list)

        if (
            leaf_num != 1
            or len(opcodes) < self.min_nb_opcode
            or len(cst_leaf_values) < self.min_nb_constant
        ):
            return None

        if logger.debug_on:
            logger.debug("Found candidate: %s", format_minsn_t(instruction))
        try:
            val_0 = tmp.evaluate_with_leaf_info(leaf_info_list, [0])  # * leaf_num)
            val_1 = tmp.evaluate_with_leaf_info(
                leaf_info_list, [0xFFFFFFFF]
            )  # * leaf_num)
            if logger.debug_on:
                logger.debug("  val_0: %s, val_1: %s", val_0, val_1)
            if val_0 != val_1 or tmp.mop is None:
                return None

            # TODO(w00tzenheimer): if we're evaluating (evaluate_with_leaf_info) and the results are equal,
            #   why do we need to run the z3 equality check?
            #   why can't this simply be:
            #   if val_0 != val_1 or tmp.mop is None:
            #       return None
            #   tmp.add_constant_leaf("c_res", val_0, tmp.mop.size)
            #   tmp.compute_sub_ast()
            #   new_instruction = self.get_replacement(typing.cast(AstNode, tmp))
            #   return new_instruction
            c_res_mop = ida_hexrays.mop_t()
            c_res_mop.make_number(val_0, tmp.mop.size or 1)
            if z3_check_mop_equality(tmp.mop, c_res_mop):
                if logger.debug_on:
                    logger.debug("  z3_check_mop_equality is equal")

                tmp.add_constant_leaf("c_res", val_0, tmp.mop.size)
                # TODO(w00tzenheimer): should we recompute caches so that leafs_by_name contains the new constant leaf?
                # tmp.compute_sub_ast()
                candidate_ast = tmp._target if isinstance(tmp, AstProxy) else tmp
                new_instruction = self.get_replacement(
                    typing.cast(AstNode, candidate_ast)
                )
                return new_instruction
        except ZeroDivisionError:
            logger.error("ZeroDivisionError while evaluating %s", tmp, exc_info=True)
        except AstEvaluationException as e:
            logger.error("Error while evaluating %s: %s", tmp, e, exc_info=True)

    @typing.override
    def check_candidate(self, candidate: AstNode) -> bool:
        """Return True if the candidate matches the rule, otherwise False."""
        return True



"""Match context for ctree pattern matching.

Stores named variable bindings during pattern matching. When a pattern
with a ``bind_name`` matches, the matched ``citem_t`` is stored under
that key for later retrieval.

Ported from herast (herast/tree/match_context.py).
"""
from __future__ import annotations

from d810.core import typing
from d810.core import getLogger

if typing.TYPE_CHECKING:
    from d810.ctree.patterns.base_pattern import BasePat

logger = getLogger("D810.ctree")

# ---------------------------------------------------------------------------
# IDA imports are optional so the module can be tested without IDA.
# ---------------------------------------------------------------------------
try:
    import idaapi
except ImportError:
    idaapi = None  # type: ignore[assignment]


class MatchContext:
    """Dict-like storage for named variable bindings during pattern matching."""

    def __init__(self, ast_ctx: typing.Any, pattern: "BasePat") -> None:
        self.ast_ctx = ast_ctx
        self.pattern = pattern
        self.binded_items: dict[str, typing.Any] = {}

    def get_item(self, name: str) -> typing.Any | None:
        """Return the item bound to *name*, or ``None``."""
        return self.binded_items.get(name, None)

    def bind_item(self, name: str, item: typing.Any) -> bool:
        """Bind *item* to *name*.

        If *name* is already bound, the new item must be equivalent to the
        existing one (same variable index for ``cot_var``, or
        ``equal_effect`` for everything else).
        """
        current_item = self.get_item(name)
        if current_item is None:
            self.binded_items[name] = item
            return True

        if idaapi is not None:
            if current_item.op == idaapi.cot_var:
                if item.op != idaapi.cot_var:
                    return False
                return current_item.v.idx == item.v.idx

        return item.equal_effect(current_item)

    def has_item(self, name: str) -> bool:
        """Return ``True`` if *name* is bound."""
        return self.binded_items.get(name, None) is not None
